from __future__ import print_function
from math import log

import numpy as np
from numpy.linalg import inv, solve
from ase.units import Bohr, Hartree

from gpaw import GPAW
from gpaw.external import ConstantElectricField
from gpaw.utilities.blas import gemm
from gpaw.mixer import DummyMixer
from gpaw.tddft.units import attosec_to_autime, autime_to_attosec
from gpaw.xc import XC

from gpaw.utilities.scalapack import (pblas_simple_hemm, pblas_simple_gemm,
                                      scalapack_inverse, scalapack_solve,
                                      scalapack_zero, pblas_tran,
                                      scalapack_set)
                                     
from time import localtime


class KickHamiltonian:
    def __init__(self, calc, ext):
        ham = calc.hamiltonian
        dens = calc.density
        vext_g = ext.get_potential(ham.finegd)
        self.vt_sG = [ham.restrict_and_collect(vext_g)]
        self.dH_asp = ham.setups.empty_atomic_matrix(1, ham.atom_partition)

        W_aL = dens.ghat.dict()
        dens.ghat.integrate(vext_g, W_aL)
        # XXX this is a quick hack to get the distribution right
        dHtmp_asp = ham.atomic_matrix_distributor.distribute(self.dH_asp)
        for a, W_L in W_aL.items():
            setup = dens.setups[a]
            dHtmp_asp[a] = np.dot(setup.Delta_pL, W_L).reshape((1, -1))
        self.dH_asp = ham.atomic_matrix_distributor.collect(dHtmp_asp)

class LCAOTDDFT(GPAW):
    def __init__(self, filename=None,
                 propagator='cn', fxc=None, **kwargs):
        self.time = 0.0
        self.niter = 0
        self.kick_strength = np.array([0.0, 0.0, 0.0], dtype=float)
        GPAW.__init__(self, filename, **kwargs)
        self.tddft_initialized = False
        self.fxc = fxc
        self.propagator = propagator

        # Restarting from a file
        if filename is not None:
            self.initialize()
            self.set_positions()

    def set(self, **kwargs):
        if 'mode' not in kwargs:
            kwargs['mode'] = 'lcao'
        else:
            modevar = kwargs['mode']
            if hasattr(modevar, 'name'):
                assert modevar.name == 'lcao'  # LCAO mode required
            else:
                assert modevar == 'lcao'  # LCAO mode required
        GPAW.set(self, **kwargs)

    def propagate_wfs(self, sourceC_nm, targetC_nm, S_MM, H_MM, dt):
        if self.propagator == 'cn':
            return self.linear_propagator(sourceC_nm, targetC_nm, S_MM, H_MM,
                                          dt)
        raise NotImplementedError

    def linear_propagator(self, sourceC_nM, targetC_nM, S_MM, H_MM, dt):
        self.timer.start('Linear solve')

        if self.blacs:
            # XXX, Preallocate
            target_blockC_nm = self.Cnm_block_descriptor.empty(dtype=complex)
            temp_blockC_nm = self.Cnm_block_descriptor.empty(dtype=complex)
            temp_block_mm = self.mm_block_descriptor.empty(dtype=complex)
            if self.density.gd.comm.rank != 0:
                # XXX Fake blacks nbands, nao, nbands, nao grid because some
                # weird asserts
                # (these are 0,x or x,0 arrays)
                sourceC_nM = self.CnM_unique_descriptor.zeros(dtype=complex)

            # 1. target = (S+0.5j*H*dt) * source
            # Wave functions to target
            self.CnM2nm.redistribute(sourceC_nM, temp_blockC_nm)

            # XXX It can't be this f'n hard to symmetrize a matrix (tri2full)
            # Remove upper diagonal
            scalapack_zero(self.mm_block_descriptor, H_MM, 'U')
            # Lower diagonal matrix:
            temp_block_mm[:] = S_MM - (0.5j * dt) * H_MM
            scalapack_set(self.mm_block_descriptor, temp_block_mm, 0, 0, 'U')
            # Note it's strictly lower diagonal matrix
            # Add transpose of H
            pblas_tran(-0.5j * dt, H_MM, 1.0, temp_block_mm,
                       self.mm_block_descriptor, self.mm_block_descriptor)
            # Add transpose of S
            pblas_tran(1.0, S_MM, 1.0, temp_block_mm,
                       self.mm_block_descriptor, self.mm_block_descriptor)

            pblas_simple_gemm(self.Cnm_block_descriptor,
                              self.mm_block_descriptor,
                              self.Cnm_block_descriptor,
                              temp_blockC_nm,
                              temp_block_mm,
                              target_blockC_nm)
            # 2. target = (S-0.5j*H*dt)^-1 * target
            # temp_block_mm[:] = S_MM + (0.5j*dt) * H_MM
            # XXX It can't be this f'n hard to symmetrize a matrix (tri2full)
            # Lower diagonal matrix:
            temp_block_mm[:] = S_MM + (0.5j * dt) * H_MM
            # Not it's stricly lower diagonal matrix:
            scalapack_set(self.mm_block_descriptor, temp_block_mm, 0, 0, 'U')
            # Add transpose of H:
            pblas_tran(+0.5j * dt, H_MM, 1.0, temp_block_mm,
                       self.mm_block_descriptor, self.mm_block_descriptor)
            # Add transpose of S
            pblas_tran(1.0, S_MM, 1.0, temp_block_mm,
                       self.mm_block_descriptor, self.mm_block_descriptor)

            scalapack_solve(self.mm_block_descriptor,
                            self.Cnm_block_descriptor,
                            temp_block_mm,
                            target_blockC_nm)

            if self.density.gd.comm.rank != 0:  # XXX is this correct?
                # XXX Fake blacks nbands, nao, nbands, nao grid because some
                # weird asserts
                # (these are 0,x or x,0 arrays)
                target = self.CnM_unique_descriptor.zeros(dtype=complex)
            else:
                target = targetC_nM
            self.Cnm2nM.redistribute(target_blockC_nm, target)
            self.density.gd.comm.broadcast(targetC_nM, 0)  # Is this required?
        else:
            # Note: The full equation is conjugated (therefore -+, not +-)
            targetC_nM[:] = \
                solve(S_MM - 0.5j * H_MM * dt,
                      np.dot(S_MM + 0.5j * H_MM * dt,
                             sourceC_nM.T.conjugate())).T.conjugate()
        
        self.timer.stop('Linear solve')

    def taylor_propagator(self, sourceC_nM, targetC_nM, S_MM, H_MM, dt):
        self.timer.start('Taylor propagator')

        if self.blacs:
            # XXX, Preallocate
            target_blockC_nm = self.Cnm_block_descriptor.empty(dtype=complex)
            if self.density.gd.comm.rank != 0:
                # XXX Fake blacks nbands, nao, nbands, nao grid because some
                # weird asserts
                # (these are 0,x or x,0 arrays)
                sourceC_nM = self.CnM_unique_descriptor.zeros(dtype=complex)

            # Zeroth order taylor to target
            self.CnM2nm.redistribute(sourceC_nM, target_blockC_nm)

            # XXX, preallocate, optimize use of temporal arrays
            temp_blockC_nm = target_blockC_nm.copy()
            temp2_blockC_nm = target_blockC_nm.copy()

            order = 4
            assert self.wfs.kd.comm.size == 1
            for n in range(order):
                # Multiply with hamiltonian
                pblas_simple_hemm(self.mm_block_descriptor,
                                  self.Cnm_block_descriptor,
                                  self.Cnm_block_descriptor,
                                  H_MM,
                                  temp_blockC_nm,
                                  temp2_blockC_nm, side='R')
                # XXX: replace with not simple gemm
                temp2_blockC_nm *= -1j * dt / (n + 1)
                # Multiply with inverse overlap
                pblas_simple_hemm(self.mm_block_descriptor,
                                  self.Cnm_block_descriptor,
                                  self.Cnm_block_descriptor,
                                  self.wfs.kpt_u[0].invS_MM,  # XXX
                                  temp2_blockC_nm,
                                  temp_blockC_nm, side='R')
                target_blockC_nm += temp_blockC_nm
            if self.density.gd.comm.rank != 0:  # Todo: Change to gd.rank
                # XXX Fake blacks nbands, nao, nbands, nao grid because
                # some weird asserts
                # (these are 0,x or x,0 arrays)
                target = self.CnM_unique_descriptor.zeros(dtype=complex)
            else:
                target = targetC_nM
            self.Cnm2nM.redistribute(target_blockC_nm, target)

            self.density.gd.comm.broadcast(targetC_nM, 0)
        else:
            assert self.wfs.kd.comm.size == 1
            if self.density.gd.comm.rank == 0:
                targetC_nM[:] = sourceC_nM[:]
                tempC_nM = sourceC_nM.copy()
                order = 4
                for n in range(order):
                    tempC_nM[:] = \
                        np.dot(self.wfs.kpt_u[0].invS,
                               np.dot(H_MM, 1j * dt / (n + 1)
                                      * tempC_nM.T.conjugate())).T.conjugate()
                    targetC_nM += tempC_nM
            self.density.gd.comm.broadcast(targetC_nM, 0)
                
        self.timer.stop('Taylor propagator')

    def absorption_kick(self, kick_strength):
        self.tddft_init()
        self.timer.start('Kick')
        self.kick_strength = np.array(kick_strength, dtype=float)

        # magnitude
        magnitude = np.sqrt(np.sum(self.kick_strength**2))

        # normalize
        direction = self.kick_strength / magnitude

        self.text('Applying absorption kick')
        self.text('Magnitude: %.8f hartree/bohr' % magnitude)
        self.text('Direction: %.4f %.4f %.4f' % tuple(direction))

        # Create hamiltonian object for absorption kick
        cef = ConstantElectricField(magnitude * Hartree / Bohr, direction)
        kick_hamiltonian = KickHamiltonian(self, cef)
        for k, kpt in enumerate(self.wfs.kpt_u):
            Vkick_MM = self.wfs.eigensolver.calculate_hamiltonian_matrix(
                kick_hamiltonian, self.wfs, kpt, add_kinetic=False, root=-1)
            for i in range(10):
                self.propagate_wfs(kpt.C_nM, kpt.C_nM, kpt.S_MM, Vkick_MM, 0.1)
        self.timer.stop('Kick')

    def blacs_mm_to_global(self, H_mm):
        # Someone could verify that this works and remove the error.
        raise NotImplementedError('Method untested and thus unreliable')
        target = self.MM_descriptor.empty(dtype=complex)
        self.mm2MM.redistribute(H_mm, target)
        self.wfs.world.barrier()
        return target

    def blacs_nm_to_global(self, C_nm):
        # Someone could verify that this works and remove the error.
        raise NotImplementedError('Method untested and thus unreliable')
        target = self.CnM_unique_descriptor.empty(dtype=complex)
        self.Cnm2nM.redistribute(C_nm, target)
        self.wfs.world.barrier()
        return target

    def tddft_init(self):
        if self.tddft_initialized:
            return
        self.blacs = self.wfs.ksl.using_blacs
        if self.blacs:
            self.ksl = ksl = self.wfs.ksl
            nao = ksl.nao
            nbands = ksl.bd.nbands
            mynbands = ksl.bd.mynbands
            blocksize = ksl.blocksize

            from gpaw.blacs import Redistributor
            if self.wfs.world.rank == 0:
                print('BLACS Parallelization')

            # Parallel grid descriptors
            grid = ksl.blockgrid
            assert grid.nprow * grid.npcol == self.wfs.ksl.block_comm.size
            # FOR DEBUG
            self.MM_descriptor = grid.new_descriptor(nao, nao, nao, nao)
            self.mm_block_descriptor = grid.new_descriptor(nao, nao, blocksize,
                                                           blocksize)
            self.Cnm_block_descriptor = grid.new_descriptor(nbands, nao,
                                                            blocksize,
                                                            blocksize)
            #self.CnM_descriptor = ksl.blockgrid.new_descriptor(nbands,
            #    nao, mynbands, nao)
            self.mM_column_descriptor = ksl.single_column_grid.new_descriptor(
                nao, nao, ksl.naoblocksize, nao)
            self.CnM_unique_descriptor = ksl.single_column_grid.new_descriptor(
                nbands, nao, mynbands, nao)

            # Redistributors
            self.mm2MM = Redistributor(ksl.block_comm,
                                       self.mm_block_descriptor,
                                       self.MM_descriptor) # XXX FOR DEBUG
            self.MM2mm = Redistributor(ksl.block_comm,
                                       self.MM_descriptor,
                                       self.mm_block_descriptor) # XXX FOR DEBUG
            self.Cnm2nM = Redistributor(ksl.block_comm,
                                        self.Cnm_block_descriptor,
                                        self.CnM_unique_descriptor)
            self.CnM2nm = Redistributor(ksl.block_comm,
                                        self.CnM_unique_descriptor,
                                        self.Cnm_block_descriptor)
            self.mM2mm = Redistributor(ksl.block_comm,
                                       self.mM_column_descriptor,
                                       self.mm_block_descriptor)

            for kpt in self.wfs.kpt_u:
                scalapack_zero(self.mm_block_descriptor, kpt.S_MM, 'U')
                scalapack_zero(self.mm_block_descriptor, kpt.T_MM, 'U')

            # XXX to propagator class
            if self.propagator == 'taylor' and self.blacs:
                # cholS_mm = self.mm_block_descriptor.empty(dtype=complex)
                for kpt in self.wfs.kpt_u:
                    kpt.invS_MM = kpt.S_MM.copy()
                    scalapack_inverse(self.mm_block_descriptor,
                                      kpt.invS_MM, 'L')
            if self.propagator == 'taylor' and not self.blacs:
                tmp = inv(self.wfs.kpt_u[0].S_MM)
                self.wfs.kpt_u[0].invS = tmp

        # Reset the density mixer
        self.density.mixer = DummyMixer()
        self.tddft_initialized = True
        for k, kpt in enumerate(self.wfs.kpt_u):
            kpt.C2_nM = kpt.C_nM.copy()
            #kpt.firstC_nM = kpt.C_nM.copy()

    def update_projectors(self):
        self.timer.start('LCAO update projectors')
        # Loop over all k-points
        for k, kpt in enumerate(self.wfs.kpt_u):
            for a, P_ni in kpt.P_ani.items():
                P_ni.fill(117)
                gemm(1.0, self.wfs.P_aqMi[a][kpt.q], kpt.C_nM, 0.0, P_ni, 'n')
        self.timer.stop('LCAO update projectors')

    def get_hamiltonian(self, kpt):
        eig = self.wfs.eigensolver
        H_MM = eig.calculate_hamiltonian_matrix(self.hamiltonian, self.wfs,
                                                kpt, root=-1)
        return H_MM

    def save_wfs(self):
        for k, kpt in enumerate(self.wfs.kpt_u):
            kpt.C2_nM[:] = kpt.C_nM

    def update_hamiltonian(self):
        self.update_projectors()
        self.density.update(self.wfs)
        self.hamiltonian.update(self.density)

    def propagate_single(self, dt):
        # --------------
        # Predictor step
        # --------------
        # 1. Calculate H(t)
        self.save_wfs() # kpt.C2_nM = kpt.C_nM
        # 2. H_MM(t) = <M|H(t)|H>
        #    Solve Psi(t+dt) from (S_MM - 0.5j*H_MM(t)*dt) Psi(t+dt) =
        #                              (S_MM + 0.5j*H_MM(t)*dt) Psi(t)

        for k, kpt in enumerate(self.wfs.kpt_u):
            if self.fxc is not None:
                if self.time == 0.0:
                    kpt.deltaXC_H_MM = self.get_hamiltonian(kpt)
                    self.hamiltonian.xc = XC(self.fxc)
                    self.update_hamiltonian()
                    assert len(self.wfs.kpt_u) == 1
                    kpt.deltaXC_H_MM -= self.get_hamiltonian(kpt)

        self.update_hamiltonian()

        # Call registered callback functions
        self.call_observers(self.niter)

        for k, kpt in enumerate(self.wfs.kpt_u):
            kpt.H0_MM = self.get_hamiltonian(kpt)
            if self.fxc is not None:
                kpt.H0_MM += kpt.deltaXC_H_MM
            self.propagate_wfs(kpt.C_nM, kpt.C_nM, kpt.S_MM, kpt.H0_MM, dt)
        # ---------------
        # Propagator step
        # ---------------
        # 1. Calculate H(t+dt)
        self.update_hamiltonian()
        # 2. Estimate H(t+0.5*dt) ~ H(t) + H(t+dT)
        for k, kpt in enumerate(self.wfs.kpt_u):
            kpt.H0_MM *= 0.5
            if self.fxc is not None:
                #  Store this to H0_MM and maybe save one extra H_MM of
                # memory?
                kpt.H0_MM += 0.5 * (self.get_hamiltonian(kpt)
                                    + kpt.deltaXC_H_MM)
            else:
                #  Store this to H0_MM and maybe save one extra H_MM of
                # memory?
                kpt.H0_MM += 0.5 * self.get_hamiltonian(kpt)

            # 3. Solve Psi(t+dt) from
            # (S_MM - 0.5j*H_MM(t+0.5*dt)*dt) Psi(t+dt)
            #    = (S_MM + 0.5j*H_MM(t+0.5*dt)*dt) Psi(t)
            self.propagate_wfs(kpt.C2_nM, kpt.C_nM, kpt.S_MM, kpt.H0_MM, dt)
        self.niter += 1
        self.time += dt

    def propagate(self, time_step=10, iterations=2000, out='lcao.dm',
                  dump_interval=50):
        assert self.wfs.dtype == complex
        assert len(self.wfs.kpt_u) == 1

        time_step *= attosec_to_autime
        self.time_step = time_step
        self.dump_interval = dump_interval
        self.tdmaxiter = self.niter + iterations

        if self.wfs.world.rank == 0:
            if self.time < self.time_step:
                self.dm_file = open(out, 'w')

                header = '# Kick = [%22.12le, %22.12le, %22.12le]\n' \
                       % (self.kick_strength[0], self.kick_strength[1], \
                          self.kick_strength[2])
                header += '# %15s %15s %22s %22s %22s\n' \
                       % ('time', 'norm', 'dmx', 'dmy', 'dmz')
                self.dm_file.write(header)
                self.dm_file.flush()
                self.text('About to do %d propagation steps.' % iterations)
            else:
                self.dm_file = open(out, 'a')
                self.text('About to continue from iteration %d and do %d '
                          'propagation steps' % (self.niter, self.tdmaxiter))
        self.tddft_init()

        dm0 = None  # Initial dipole moment
        self.timer.start('Propagate')
        while self.niter < self.tdmaxiter:
            dm = self.density.finegd.calculate_dipole_moment(
                self.density.rhot_g)
            if dm0 is None:
                dm0 = dm

            norm = self.density.finegd.integrate(self.density.rhot_g)
            line = ('%20.8lf %20.8le %22.12le %22.12le %22.12le'
                    % (self.time, norm, dm[0], dm[1], dm[2]))
            T = localtime()
            if self.wfs.world.rank == 0:
                print(line, file=self.dm_file)

            if self.wfs.world.rank == 0 and self.niter % 1 == 0:
                self.text('iter: %3d  %02d:%02d:%02d %11.2f   %9.1f'
                          % (self.niter, T[3], T[4], T[5],
                             self.time * autime_to_attosec,
                             log(abs(norm) + 1e-16) / log(10)))
                self.dm_file.flush()
            self.propagate_single(self.time_step)
            
        self.call_observers(self.niter, final=True)
        if self.wfs.world.rank == 0:
            self.dm_file.close()
        self.timer.stop('Propagate')




import sys
import numpy as np

from ase.utils import devnull

from gpaw import debug
from gpaw.mpi import world
from gpaw.utilities.tools import tri2full
from gpaw.hs_operators import MatrixOperator
from gpaw.utilities import compiled_with_sl
from gpaw.utilities.scalapack import scalapack_set
from gpaw.blacs import Redistributor
from gpaw.kohnsham_layouts import BlacsBandLayouts
if 0:  # causes numpy doctests failures - exact formatting is expected!
    np.set_printoptions(linewidth=168) #XXX large xterm width

memstats = False

from gpaw.test.ut_common import TextTestRunner, \
    CustomTextTestRunner, defaultTestLoader, initialTestLoader

from gpaw.test.parallel.ut_hsops import UTBandParallelSetup, \
                                        UTConstantWavefunctionSetup

# -------------------------------------------------------------------

class UTBandParallelBlacsSetup(UTBandParallelSetup):
    """
    Setup a simple band parallel calculation using BLACS."""

    def create_kohn_sham_layouts(self):
        # Find BLACS parameters for Kohn-Sham layouts
        cpus = self.bd.comm.size * self.gd.comm.size
        self.mcpus = int(cpus**0.5)
        self.ncpus = cpus//self.mcpus
        return BlacsBandLayouts(self.gd, self.bd, self.block_comm, self.dtype, self.mcpus, self.ncpus, 6)

    # =================================

    def verify_comm_sizes(self):
        if world.size == 1:
            return
        comm_sizes = tuple([comm.size for comm in [world, self.bd.comm, \
                                                   self.gd.comm, self.kpt_comm]])
        comm_sizes += (self.mcpus, self.ncpus)
        self._parinfo =  '%d world, %d band, %d domain, %d kpt, %d x %d BLACS' % comm_sizes
        self.assertEqual((self.nspins*self.nibzkpts) % self.kpt_comm.size, 0)

    def verify_kohn_sham_layouts(self):
        # TODO do more here :)
        self.assertTrue(self.ksl.using_blacs)
        self.assertTrue(self.ksl.bd is self.bd)
        self.assertTrue(self.ksl.gd is self.gd)


class UTBandParallelBlacsSetup_Blocked(UTBandParallelBlacsSetup):
    __doc__ = UTBandParallelBlacsSetup.__doc__
    parstride_bands = False
    dtype = float

class UTBandParallelBlacsSetup_Strided(UTBandParallelSetup):
    __doc__ = UTBandParallelBlacsSetup.__doc__
    parstride_bands = True
    dtype = float

# -------------------------------------------------------------------

class UTConstantWavefunctionBlacsSetup(UTConstantWavefunctionSetup,
                                       UTBandParallelBlacsSetup):
    __doc__ = UTBandParallelBlacsSetup.__doc__ + """
    The pseudo wavefunctions are constants normalized to their band index."""

    # =================================

    def test_overlaps_hermitian(self):
        # Set up Hermitian overlap operator:
        S = lambda x: x
        dS = lambda a, P_ni: np.dot(P_ni, self.setups[a].dO_ii)
        nblocks = self.get_optimal_number_of_blocks(self.blocking)
        overlap = MatrixOperator(self.ksl, nblocks, self.async, True)
        S_nn = overlap.calculate_matrix_elements(self.psit_nG, self.P_ani, S, dS)

        if memstats:
            self.mem_test = record_memory()

        S_NN = self.ksl.nndescriptor.collect_on_master(S_nn)
        if self.bd.comm.rank == 0 and self.gd.comm.rank == 0:
            assert S_NN.shape == (self.bd.nbands,) * 2
            S_NN = S_NN.T.copy() # Fortran -> C indexing
            tri2full(S_NN, 'U') # upper to lower...
        else:
            assert S_NN.nbytes == 0
            S_NN = np.empty((self.bd.nbands,) * 2, dtype=S_NN.dtype)

        if self.bd.comm.rank == 0:
            self.gd.comm.broadcast(S_NN, 0)
        self.bd.comm.broadcast(S_NN, 0)

        self.check_and_plot(S_NN, self.S0_nn, 9, 'overlaps,hermitian')

    def test_overlaps_nonhermitian(self):
        alpha = np.random.normal(size=1).astype(self.dtype)
        if self.dtype == complex:
            alpha += 1j*np.random.normal(size=1)
        world.broadcast(alpha, 0)

        # Set up non-Hermitian overlap operator:
        S = lambda x: alpha*x
        dS = lambda a, P_ni: np.dot(alpha*P_ni, self.setups[a].dO_ii)
        nblocks = self.get_optimal_number_of_blocks(self.blocking)
        overlap = MatrixOperator(self.ksl, nblocks, self.async, False)
        if 0: #XXX non-hermitian case so Nn2nn not just uplo='L' but rather 'G'
            blockcomm = self.ksl.nndescriptor.blacsgrid.comm
            self.ksl.Nn2nn = Redistributor(blockcomm, self.ksl.Nndescriptor,
                                           self.ksl.nndescriptor)
        S_nn = overlap.calculate_matrix_elements(self.psit_nG, self.P_ani, S, dS)

        if memstats:
            self.mem_test = record_memory()

        S_NN = self.ksl.nndescriptor.collect_on_master(S_nn)
        if self.bd.comm.rank == 0 and self.gd.comm.rank == 0:
            assert S_NN.shape == (self.bd.nbands,) * 2
            S_NN = S_NN.T.copy() # Fortran -> C indexing
        else:
            assert S_NN.nbytes == 0
            S_NN = np.empty((self.bd.nbands,) * 2, dtype=S_NN.dtype)

        if self.bd.comm.rank == 0:
            self.gd.comm.broadcast(S_NN, 0)
        self.bd.comm.broadcast(S_NN, 0)

        self.check_and_plot(S_NN, alpha*self.S0_nn, 9, 'overlaps,nonhermitian')

    def test_trivial_cholesky(self):
        # Set up Hermitian overlap operator:
        S = lambda x: x
        dS = lambda a, P_ni: np.dot(P_ni, self.setups[a].dO_ii)
        nblocks = self.get_optimal_number_of_blocks(self.blocking)
        overlap = MatrixOperator(self.ksl, nblocks, self.async, True)
        S_nn = overlap.calculate_matrix_elements(self.psit_nG, self.P_ani, S, dS)

        # Known starting point of SI_nn = <psit_m|S+alpha*I|psit_n>
        I_nn = self.ksl.nndescriptor.empty(dtype=S_nn.dtype)
        scalapack_set(self.ksl.nndescriptor, I_nn, 0.0, 1.0, 'L')
        alpha = 1e-3 # shift eigenvalues away from zero

        C_nn = S_nn + alpha * I_nn
        self.ksl.nndescriptor.inverse_cholesky(C_nn, 'L')
        self.psit_nG = overlap.matrix_multiply(C_nn, self.psit_nG, self.P_ani)
        D_nn = overlap.calculate_matrix_elements(self.psit_nG, self.P_ani, S, dS)

        D_NN = self.ksl.nndescriptor.collect_on_master(D_nn)
        if self.bd.comm.rank == 0 and self.gd.comm.rank == 0:
            assert D_NN.shape == (self.bd.nbands,) * 2
            D_NN = D_NN.T.copy() # Fortran -> C indexing
            tri2full(D_NN, 'U') # upper to lower..
        else:
            assert D_NN.nbytes == 0
            D_NN = np.empty((self.bd.nbands,) * 2, dtype=D_NN.dtype)

        if self.bd.comm.rank == 0:
            self.gd.comm.broadcast(D_NN, 0)
        self.bd.comm.broadcast(D_NN, 0)

        # D_NN = C_NN^dag * S_NN * C_NN = I_NN - alpha * C_NN^dag * C_NN
        I_NN = np.eye(self.bd.nbands)
        C0_NN = np.linalg.inv(np.linalg.cholesky(self.S0_nn + alpha*I_NN).T.conj())
        D0_NN = I_NN - alpha * np.dot(C0_NN.T.conj(), C0_NN)
        self.check_and_plot(D_NN, D0_NN, 6, 'trivial,cholesky') #XXX precision

    def test_trivial_diagonalize(self): #XXX XXX XXX
        # Known starting point of S_nn = <psit_m|S|psit_n>
        S_nn = self.S0_nn

        # Eigenvector decomposition S_nn = V_nn * W_nn * V_nn^dag
        # Utilize the fact that they are analytically known (cf. Maple)
        W_n = np.zeros(self.nbands).astype(self.dtype)
        W_n[1] = (1. + self.Qtotal) * self.nbands * (self.nbands - 1) / 2.

        # Set up Hermitian overlap operator:
        S = lambda x: x
        dS = lambda a, P_ni: np.dot(P_ni, self.setups[a].dO_ii)
        nblocks = self.get_optimal_number_of_blocks(self.blocking)
        overlap = MatrixOperator(self.ksl, nblocks, self.async, True)
        S_nn = overlap.calculate_matrix_elements(self.psit_nG, self.P_ani, S, dS)

        eps_N = self.bd.empty(global_array=True) # XXX dtype?
        C_nn = self.ksl.nndescriptor.empty(dtype=S_nn.dtype)
        self.ksl.nndescriptor.diagonalize_dc(S_nn, C_nn, eps_N, 'L')
        self.assertAlmostEqual(np.abs(np.sort(eps_N)-np.sort(W_n)).max(), 0, 9)

        #eps_n = self.bd.empty()
        #self.bd.distribute(eps_N, eps_n) # XXX only blocked groups, right?

        # Rotate wavefunctions to diagonalize the overlap
        self.psit_nG = overlap.matrix_multiply(C_nn, self.psit_nG, self.P_ani)

        # Recaulculate the overlap matrix, which should now be diagonal
        D_nn = overlap.calculate_matrix_elements(self.psit_nG, self.P_ani, S, dS)

        D_NN = self.ksl.nndescriptor.collect_on_master(D_nn)
        if self.bd.comm.rank == 0 and self.gd.comm.rank == 0:
            assert D_NN.shape == (self.bd.nbands,) * 2
            D_NN = D_NN.T.copy() # Fortran -> C indexing
            tri2full(D_NN, 'U') # upper to lower...
        else:
            assert D_NN.nbytes == 0
            D_NN = np.empty((self.bd.nbands,) * 2, dtype=D_NN.dtype)

        if self.bd.comm.rank == 0:
            self.gd.comm.broadcast(D_NN, 0)
        self.bd.comm.broadcast(D_NN, 0)

        D0_NN = np.diag(eps_N)
        self.check_and_plot(D_NN, D0_NN, 9, 'trivial,diagonalize')

    def test_multiply_randomized(self):
        # Known starting point of S_nn = <psit_m|S|psit_n>
        S_NN = self.S0_nn

        if self.dtype == complex:
            C_NN = np.random.uniform(size=self.nbands**2) * \
                np.exp(1j*np.random.uniform(0,2*np.pi,size=self.nbands**2))
        else:
            C_NN = np.random.normal(size=self.nbands**2)
        C_NN = C_NN.reshape((self.nbands,self.nbands)) / np.linalg.norm(C_NN,2)
        world.broadcast(C_NN, 0)

        # Set up Hermitian overlap operator:
        S = lambda x: x
        dS = lambda a, P_ni: np.dot(P_ni, self.setups[a].dO_ii)
        nblocks = self.get_optimal_number_of_blocks(self.blocking)
        overlap = MatrixOperator(self.ksl, nblocks, self.async, True)

        if self.bd.comm.rank == 0 and self.gd.comm.rank == 0:
            assert C_NN.shape == (self.bd.nbands,) * 2
            tmp_NN = C_NN.T.copy() # C -> Fortran indexing
        else:
            tmp_NN = self.ksl.nndescriptor.as_serial().empty(dtype=C_NN.dtype)
        C_nn = self.ksl.nndescriptor.distribute_from_master(tmp_NN)

        self.psit_nG = overlap.matrix_multiply(C_nn, self.psit_nG, self.P_ani)
        D_nn = overlap.calculate_matrix_elements(self.psit_nG, self.P_ani, S, dS)

        if memstats:
            self.mem_test = record_memory()

        D_NN = self.ksl.nndescriptor.collect_on_master(D_nn)
        if self.bd.comm.rank == 0 and self.gd.comm.rank == 0:
            assert D_NN.shape == (self.bd.nbands,) * 2
            D_NN = D_NN.T.copy() # Fortran -> C indexing
            tri2full(D_NN, 'U') # upper to lower...
        else:
            assert D_NN.nbytes == 0
            D_NN = np.empty((self.bd.nbands,) * 2, dtype=D_NN.dtype)

        if self.bd.comm.rank == 0:
            self.gd.comm.broadcast(D_NN, 0)
        self.bd.comm.broadcast(D_NN, 0)

        # D_nn = C_nn^dag * S_nn * C_nn
        D0_NN = np.dot(C_NN.T.conj(), np.dot(S_NN, C_NN))
        self.check_and_plot(D_NN, D0_NN, 9, 'multiply,randomized')

    def test_multiply_nonhermitian(self):
        alpha = np.random.normal(size=1).astype(self.dtype)
        if self.dtype == complex:
            alpha += 1j*np.random.normal(size=1)
        world.broadcast(alpha, 0)

        # Known starting point of S_nn = <psit_m|S|psit_n>
        S_NN = alpha*self.S0_nn

        if self.dtype == complex:
            C_NN = np.random.uniform(size=self.nbands**2) * \
                np.exp(1j*np.random.uniform(0,2*np.pi,size=self.nbands**2))
        else:
            C_NN = np.random.normal(size=self.nbands**2)
        C_NN = C_NN.reshape((self.nbands,self.nbands)) / np.linalg.norm(C_NN,2)
        world.broadcast(C_NN, 0)

        # Set up Hermitian overlap operator:
        S = lambda x: alpha*x
        dS = lambda a, P_ni: np.dot(alpha*P_ni, self.setups[a].dO_ii)
        nblocks = self.get_optimal_number_of_blocks(self.blocking)
        overlap = MatrixOperator(self.ksl, nblocks, self.async, False)

        if 0: #XXX non-hermitian case so Nn2nn not just uplo='L' but rather 'G'
            blockcomm = self.ksl.nndescriptor.blacsgrid.comm
            self.ksl.Nn2nn = Redistributor(blockcomm, self.ksl.Nndescriptor,
                                           self.ksl.nndescriptor)

        if self.bd.comm.rank == 0 and self.gd.comm.rank == 0:
            assert C_NN.shape == (self.bd.nbands,) * 2
            tmp_NN = C_NN.T.copy() # C -> Fortran indexing
        else:
            tmp_NN = self.ksl.nndescriptor.as_serial().empty(dtype=C_NN.dtype)
        C_nn = self.ksl.nndescriptor.distribute_from_master(tmp_NN)

        self.psit_nG = overlap.matrix_multiply(C_nn, self.psit_nG, self.P_ani)
        D_nn = overlap.calculate_matrix_elements(self.psit_nG, self.P_ani, S, dS)

        if memstats:
            self.mem_test = record_memory()

        D_NN = self.ksl.nndescriptor.collect_on_master(D_nn)
        if self.bd.comm.rank == 0 and self.gd.comm.rank == 0:
            assert D_NN.shape == (self.bd.nbands,) * 2
            D_NN = D_NN.T.copy() # Fortran -> C indexing
        else:
            assert D_NN.nbytes == 0
            D_NN = np.empty((self.bd.nbands,) * 2, dtype=D_NN.dtype)

        if self.bd.comm.rank == 0:
            self.gd.comm.broadcast(D_NN, 0)
        self.bd.comm.broadcast(D_NN, 0)

        # D_nn = C_nn^dag * S_nn * C_nn
        D0_NN = np.dot(C_NN.T.conj(), np.dot(S_NN, C_NN))
        self.check_and_plot(D_NN, D0_NN, 9, 'multiply,nonhermitian')


# -------------------------------------------------------------------

def UTConstantWavefunctionFactory(dtype, parstride_bands, blocking, async):
    sep = '_'
    classname = 'UTConstantWavefunctionBlacsSetup' \
    + sep + {float:'Float', complex:'Complex'}[dtype] \
    + sep + {False:'Blocked', True:'Strided'}[parstride_bands] \
    + sep + {'fast':'Fast', 'light':'Light',
             'intdiv':'Intdiv', 'nonintdiv1': 'Nonintdiv1',
             'nonintdiv2':'Nonintdiv2'}[blocking] \
    + sep + {False:'Synchronous', True:'Asynchronous'}[async]
    class MetaPrototype(UTConstantWavefunctionBlacsSetup, object):
        __doc__ = UTConstantWavefunctionBlacsSetup.__doc__
        dtype = dtype
        parstride_bands = parstride_bands
        blocking = blocking
        async = async
    MetaPrototype.__name__ = classname
    return MetaPrototype

# -------------------------------------------------------------------

if __name__ in ['__main__', '__builtin__'] and compiled_with_sl():
    # We may have been imported by test.py, if so we should redirect to logfile
    if __name__ == '__builtin__':
        testrunner = CustomTextTestRunner('ut_hsblacs.log', verbosity=2)
    else:
        stream = (world.rank == 0) and sys.stdout or devnull
        testrunner = TextTestRunner(stream=stream, verbosity=2)

    parinfo = []
    # Initial Verification only tests case : dtype = float
    for test in [UTBandParallelBlacsSetup_Blocked]: #, UTBandParallelBlacsSetup_Strided]:
        info = ['', test.__name__, test.__doc__.strip('\n'), '']
        testsuite = initialTestLoader.loadTestsFromTestCase(test)
        list(map(testrunner.stream.writeln, info))
        testresult = testrunner.run(testsuite)
        assert testresult.wasSuccessful(), 'Initial verification failed!'
        parinfo.extend(['    Parallelization options: %s' % tci._parinfo for \
                        tci in testsuite._tests if hasattr(tci, '_parinfo')])
    parinfo = np.unique(np.sort(parinfo)).tolist()

    testcases = []
    for dtype in [float, complex]:
        for parstride_bands in [False]: #XXX [False, True]:
            for blocking in ['fast', 'light', 'intdiv',
                             'nonintdiv1', 'nonintdiv2']:
                for async in [False, True]:
                    testcases.append(UTConstantWavefunctionFactory(dtype, \
                        parstride_bands, blocking, async))

    for test in testcases:
        info = ['', test.__name__, test.__doc__.strip('\n')] + parinfo + ['']
        testsuite = defaultTestLoader.loadTestsFromTestCase(test)
        list(map(testrunner.stream.writeln, info))
        testresult = testrunner.run(testsuite)
        # Provide feedback on failed tests if imported by test.py
        if __name__ == '__builtin__' and not testresult.wasSuccessful():
            raise SystemExit('Test failed. Check ut_hsblacs.log for details.')




import os
from math import pi, cos, sin
from ase import Atom, Atoms
from ase.parallel import rank, barrier
from gpaw import GPAW, FermiDirac
from gpaw.test import equal, gen
from gpaw.xas import  * #XAS, RecursionMethod
import numpy as np
from gpaw import setup_paths
#setup_paths.insert(0, '.')


# Generate setup for oxygen with half a core-hole:
gen('Si', name='hch1s', corehole=(1, 0, 0.5))

a = 5.43095
b = a / 2
c = b / 2
d = b + c

si_nonortho = Atoms([Atom('Si', (0, 0, 0)),
            Atom('Si', (a/4, a/4, a/4))],
            cell=[(a/2,a/2,0),(a/2,0,a/2),(0,a/2,a/2)],
            pbc=True)

# calculation with full symmetry
calc = GPAW(nbands=-10,
            h=0.25,
            kpts=(2,2,2),
            occupations=FermiDirac(width=0.05),
            setups={0: 'hch1s'})

si_nonortho.set_calculator(calc)
e = si_nonortho.get_potential_energy()
niter = calc.get_number_of_iterations()
calc.write('si_nonortho_xas_sym.gpw')


# calculation without any symmetry
calc = GPAW(nbands=-10,
            h=0.25,
            kpts=(2,2,2),
            occupations=FermiDirac(width=0.05),
            setups={0: 'hch1s'},
            symmetry='off')

si_nonortho.set_calculator(calc)
e = si_nonortho.get_potential_energy()
niter = calc.get_number_of_iterations()
calc.write('si_nonortho_xas_nosym.gpw')


# restart from file
calc1 = GPAW('si_nonortho_xas_sym.gpw')
calc2 = GPAW('si_nonortho_xas_nosym.gpw')
if mpi.size == 1:
    xas1 = XAS(calc1)
    x, y1 = xas1.get_spectra()
    xas2 = XAS(calc2)
    x2, y2 = xas2.get_spectra(E_in=x)

    assert (np.sum(abs(y1-y2)[0,:500] ** 2 ) < 5e-9)
    assert (np.sum(abs(y1-y2)[1,:500] ** 2 ) < 5e-9)
    assert (np.sum(abs(y1-y2)[2,:500] ** 2 ) < 5e-9)

#else:
#    x = np.linspace(0, 10, 50)
#
#r1 = RecursionMethod(calc1)
#r1.run(40)
#
#r2 = RecursionMethod(calc1)
#r2.run(40)
#
#if mpi.size == 1:                                                                                                                                                                                                  
#    y1 = r1.get_spectra(x)
#    y2 = r2.get_spectra(x)
#
#    assert (np.sum(abs(y1-y2)[0,:] ** 2 ) < 5e-9)
#    assert (np.sum(abs(y1-y2)[1,:] ** 2 ) < 5e-9)
#    assert (np.sum(abs(y1-y2)[2,:] ** 2 ) < 5e-9)




from ase import *
from ase.structure import molecule
from gpaw import *
from gpaw.mpi import serial_comm
from gpaw.test import equal
from gpaw.xc.fxc_correlation_energy import FXCCorrelation

ecut = 50

He = Atoms('He')
He.center(vacuum=1.0)

calc = GPAW(mode='pw',
            dtype=complex,
            xc='PBE',
            communicator=serial_comm)
He.set_calculator(calc)
He.get_potential_energy()
calc.diagonalize_full_hamiltonian()

ralda = FXCCorrelation(calc, xc='rALDA')
C6_ralda, C6_0 = ralda.get_C6_coefficient(ecut=ecut,
                                          direction=2)

equal(C6_0, 1.772, 0.01)
equal(C6_ralda, 1.609, 0.01)



from __future__ import print_function

import numpy as np

from gpaw.response.g0w0 import G0W0


class GW0(G0W0):
    def __init__(self, calc, filename='gw0', **kwargs):
        G0W0.__init__(self, calc, filename, savew=True, **kwargs)
        try:
            self.qp_xsin = np.load(filename + '-gw0.npy')
        except IOError:
            self.qp_xsin = np.empty((1,) + self.shape)
            
        self.iteration = len(self.qp_xsin)
    
    def get_k_point(self, s, K, n1, n2):
        kpt = G0W0.get_k_point(self, s, K, n1, n2)
        if self.iteration > 1:
            b1, b2 = self.bands
            m1 = max(b1, n1)
            m2 = min(b2, n2)
            if m2 > m1:
                k = self.calc.wfs.kd.bz2ibz_k[K]
                i = self.kpts.index(k)
                qp_n = self.qp_xsin[-1, s, i, m1 - b1:m2 - b1]
                kpt.eps_n[m1 - n1:m2 - n1] = qp_n
                
        return kpt
    
    def update_qp_energies(self):
        print('Iteration:', self.iteration, file=self.fd)
        
        if self.iteration == 1:
            self.qp_xsin[0] = self.eps_sin
            
        self.Z_sin = 1 / (1 - self.dsigma_sin)
        self.qp_sin = self.eps_sin + self.Z_sin * (
            self.sigma_sin + self.exx_sin -
            self.vxc_sin)
        
        self.iteration += 1
        qp_xsin = np.empty((self.iteration,) + self.shape)
        qp_xsin[:-1] = self.qp_xsin
        qp_xsin[-1] = self.qp_sin
        self.qp_xsin = qp_xsin
        
        if self.world.rank == 0:
            with open(self.filename + '.qp.npy', 'w') as fd:
                np.save(fd, self.qp_xsin)



import numpy as np

from ase import Atoms
from ase.units import Bohr
from gpaw.density import RealSpaceDensity
from gpaw.lfc import BasisFunctions
from gpaw.setup import Setups
from gpaw.xc import XC
from gpaw.utilities.tools import coordinates
from gpaw.utilities.partition import AtomPartition

#from gpaw.mpi import rank


class HirshfeldDensity(RealSpaceDensity):

    """Density as sum of atomic densities."""

    def __init__(self, calculator):
        self.calculator = calculator
        density = calculator.density
        par = self.calculator.input_parameters
        RealSpaceDensity.__init__(self, density.gd, density.finegd, 1, 0,
                                  stencil=par.stencils[1],
                                  grid2grid=density.grid2grid)

    def set_positions(self, spos_ac, atom_partition):
        """HirshfeldDensity builds a hack density object to calculate
        all electron density
        of atoms. This methods overrides the parallel distribution of
        atomic density matrices
        in density.py"""
        self.atom_partition = atom_partition
        self.atomic_matrix_distributor = self.grid2grid.get_matrix_distributor(
            self.atom_partition, spos_ac)
        self.nct.set_positions(spos_ac)
        self.ghat.set_positions(spos_ac)
        self.mixer.reset()
        # self.nt_sG = None
        self.nt_sg = None
        self.nt_g = None
        self.rhot_g = None
        self.Q_aL = None
        self.nct_G = self.gd.zeros()
        self.nct.add(self.nct_G, 1.0 / self.nspins)

    def get_density(self, atom_indices=None, gridrefinement=2):
        """Get sum of atomic densities from the given atom list.

        All atoms are taken if the list is not given."""

        all_atoms = self.calculator.get_atoms()
        if atom_indices is None:
            atom_indices = range(len(all_atoms))

        density = self.calculator.density
        spos_ac = all_atoms.get_scaled_positions()
        rank_a = self.finegd.get_ranks_from_positions(spos_ac)
        
        density.set_positions(all_atoms.get_scaled_positions(),
                              AtomPartition(self.finegd.comm, rank_a)
                              )

        # select atoms
        atoms = []
        D_asp = {}
        rank_a = []
        all_D_asp = self.calculator.density.D_asp
        all_rank_a = self.calculator.density.atom_partition.rank_a
        for a in atom_indices:
            if a in all_D_asp:
                D_asp[len(atoms)] = all_D_asp.get(a)
            atoms.append(all_atoms[a])
            rank_a.append(all_rank_a[a])
        atoms = Atoms(atoms,
                      cell=all_atoms.get_cell(), pbc=all_atoms.get_pbc())
        spos_ac = atoms.get_scaled_positions()
        Z_a = atoms.get_atomic_numbers()

        par = self.calculator.input_parameters
        setups = Setups(Z_a, par.setups, par.basis, par.lmax,
                        XC(par.xc),
                        self.calculator.wfs.world)

        # initialize
        self.initialize(setups,
                        self.calculator.timer,
                        np.zeros((len(atoms), 3)), False)
        self.set_mixer(None)

        # FIXME nparray causes partitionong.py test to fail
        self.set_positions(spos_ac, AtomPartition(self.gd.comm, rank_a))
        self.D_asp = D_asp
        basis_functions = BasisFunctions(self.gd,
                                         [setup.phit_j
                                          for setup in self.setups],
                                         cut=True)
        basis_functions.set_positions(spos_ac)
        self.initialize_from_atomic_densities(basis_functions)

        aed_sg, gd = self.get_all_electron_density(atoms,
                                                   gridrefinement)
        return aed_sg[0], gd


class HirshfeldPartitioning:

    """Partion space according to the Hirshfeld method.

    After: F. L. Hirshfeld Theoret. Chim.Acta 44 (1977) 129-138
    """

    def __init__(self, calculator, density_cutoff=1.e-12):
        self.calculator = calculator
        self.density_cutoff = density_cutoff

    def initialize(self):
        self.atoms = self.calculator.get_atoms()
        self.hdensity = HirshfeldDensity(self.calculator)
        density_g, gd = self.hdensity.get_density()
        self.invweight_g = 0. * density_g
        density_ok = np.where(density_g > self.density_cutoff)
        self.invweight_g[density_ok] = 1.0 / density_g[density_ok]

    def get_calculator(self):
        return self.calculator

    def get_effective_volume_ratio(self, atom_index):
        """Effective volume to free volume ratio.

        After: Tkatchenko and Scheffler PRL 102 (2009) 073005, eq. (7)
        """
        atoms = self.atoms
        finegd = self.calculator.density.finegd

        den_sg, gd = self.calculator.density.get_all_electron_density(atoms)
        den_g = den_sg.sum(axis=0)
        assert(gd == finegd)
        denfree_g, gd = self.hdensity.get_density([atom_index])
        assert(gd == finegd)

        # the atoms r^3 grid
        position = self.atoms[atom_index].position / Bohr
        r_vg, r2_g = coordinates(finegd, origin=position)
        r3_g = r2_g * np.sqrt(r2_g)

        weight_g = denfree_g * self.invweight_g

        nom = finegd.integrate(r3_g * den_g * weight_g)
        denom = finegd.integrate(r3_g * denfree_g)

        return nom / denom

    def get_weight(self, atom_index):
        denfree_g, gd = self.hdensity.get_density([atom_index])
        weight_g = denfree_g * self.invweight_g
        return weight_g

    def get_charges(self, den_g=None):
        """Charge on the atom according to the Hirshfeld partitioning

        Can be applied to any density den_g.
        """
        self.initialize()
        finegd = self.calculator.density.finegd

        if den_g is None:
            den_sg, gd = self.calculator.density.get_all_electron_density(
                self.atoms)
            den_g = den_sg.sum(axis=0)
        assert(den_g.shape == tuple(finegd.n_c))

        charges = []
        for ia, atom in enumerate(self.atoms):
            weight_g = self.get_weight(ia)
#            charge = atom.number - finegd.integrate(weight_g * den_g)
            charges.append(atom.number - finegd.integrate(weight_g * den_g))
        return charges

    def get_effective_volume_ratios(self):
        """Return the list of effective volume to free volume ratios."""
        self.initialize()
        ratios = []
        for a, atom in enumerate(self.atoms):
            ratios.append(self.get_effective_volume_ratio(a))
        return np.array(ratios)



from __future__ import division, print_function

import functools
import pickle
from math import pi

import numpy as np
from ase.dft.kpoints import monkhorst_pack
from ase.units import Hartree
from ase.utils import opencew, devnull
from ase.utils.timing import timer
from ase.parallel import paropen

import gpaw.mpi as mpi
from gpaw import debug
from gpaw.kpt_descriptor import KPointDescriptor
from gpaw.response.chi0 import Chi0, HilbertTransform
from gpaw.response.pair import PairDensity
from gpaw.response.wstc import WignerSeitzTruncatedCoulomb
from gpaw.response.kernels import get_coulomb_kernel
from gpaw.response.kernels import get_integrated_kernel
from gpaw.wavefunctions.pw import PWDescriptor, count_reciprocal_vectors
from gpaw.xc.exx import EXX, select_kpts
from gpaw.xc.tools import vxc


class G0W0(PairDensity):
    """This class defines the G0W0 calculator. The G0W0 calculator is used
    is used to calculate the quasi particle energies through the G0W0
    approximation for a number of states.

    Note: So far the G0W0 calculation only works for spin-paired systems.

    Parameters:
       calc: str or PAW object
          GPAW calculator object or filename of saved calculator object.
       filename: str
          Base filename of output files.
       kpts: list
          List of indices of the IBZ k-points to calculate the quasi particle
          energies for.
       bands: tuple
          Range of band indices, like (n1, n2+1), to calculate the quasi
          particle energies for. Note that the second band index is not
          included.
       ecut: float
          Plane wave cut-off energy in eV.
       nbands: int
          Number of bands to use in the calculation. If :None: the number will
          be determined from :ecut: to yield a number close to the number of
          plane waves used.
       ppa: bool
          Sets whether the Godby-Needs plasmon-pole approximation for the
          dielectric function should be used.
       truncation: str
            Coulomb truncation scheme. Can be either wigner-seitz,
            2D, 1D, or 0D
       integrate_gamma: int
            Method to integrate the Coulomb interaction. 1 is a numerical
            integration at all q-points with G=[0,0,0] - this breaks the
            symmetry slightly. 0 is analytical integration at q=[0,0,0] only -
            this conserves the symmetry. integrate_gamma=2 is the same as 1,
            but the average is only carried out in the non-periodic directions.
       E0: float
          Energy (in eV) used for fitting in the plasmon-pole approximation.
       domega0: float
          Minimum frequency step (in eV) used in the generation of the non-
          linear frequency grid.
       omega2: float
          Control parameter for the non-linear frequency grid, equal to the
          frequency where the grid spacing has doubled in size.
    """
    def __init__(self, calc, filename='gw',
                 kpts=None, bands=None, nbands=None, ppa=False,
                 truncation=None, integrate_gamma=1,
                 ecut=150.0, eta=0.1, E0=1.0 * Hartree,
                 domega0=0.025, omega2=10.0,
                 nblocks=1, savew=False, savepckl=True,
                 world=mpi.world):

        if world.rank != 0:
            txt = devnull
        else:
            txt = open(filename + '.txt', 'w')

        p = functools.partial(print, file=txt)
        p('  ___  _ _ _ ')
        p(' |   || | | |')
        p(' | | || | | |')
        p(' |__ ||_____|')
        p(' |___|')
        p()

        self.inputcalc = calc
        PairDensity.__init__(self, calc, ecut, world=world, nblocks=nblocks,
                             txt=txt)

        self.filename = filename
        self.savew = savew
        self.savepckl = savepckl
        
        ecut /= Hartree
        
        self.ppa = ppa
        self.truncation = truncation
        self.integrate_gamma = integrate_gamma
        self.eta = eta / Hartree
        self.E0 = E0 / Hartree
        self.domega0 = domega0 / Hartree
        self.omega2 = omega2 / Hartree

        self.kpts = list(select_kpts(kpts, self.calc))
                
        if bands is None:
            bands = [0, self.nocc2]
            
        self.bands = bands

        b1, b2 = bands
        self.shape = shape = (self.calc.wfs.nspins, len(self.kpts), b2 - b1)
        self.eps_sin = np.empty(shape)     # KS-eigenvalues
        self.f_sin = np.empty(shape)       # occupation numbers
        self.sigma_sin = np.zeros(shape)   # self-energies
        self.dsigma_sin = np.zeros(shape)  # derivatives of self-energies
        self.vxc_sin = None                # KS XC-contributions
        self.exx_sin = None                # exact exchange contributions
        self.Z_sin = None                  # renormalization factors
        
        if nbands is None:
            nbands = int(self.vol * ecut**1.5 * 2**0.5 / 3 / pi**2)
        self.nbands = nbands
        self.nspins = self.calc.wfs.nspins

        p()
        p('Quasi particle states:')
        if kpts is None:
            p('All k-points in IBZ')
        else:
            kptstxt = ', '.join(['{0:d}'.format(k) for k in self.kpts])
            p('k-points (IBZ indices): [' + kptstxt + ']')
        p('Band range: ({0:d}, {1:d})'.format(b1, b2))
        p()
        p('Computational parameters:')
        p('Plane wave cut-off: {0:g} eV'.format(self.ecut * Hartree))
        p('Number of bands: {0:d}'.format(self.nbands))
        p('Coulomb cutoff:', self.truncation)
        p('Broadening: {0:g} eV'.format(self.eta * Hartree))

        kd = self.calc.wfs.kd

        self.mysKn1n2 = None  # my (s, K, n1, n2) indices
        self.distribute_k_points_and_bands(b1, b2, kd.ibz2bz_k[self.kpts])
        
        # Find q-vectors and weights in the IBZ:
        assert -1 not in kd.bz2bz_ks
        offset_c = 0.5 * ((kd.N_c + 1) % 2) / kd.N_c
        bzq_qc = monkhorst_pack(kd.N_c) + offset_c
        self.qd = KPointDescriptor(bzq_qc)
        self.qd.set_symmetry(self.calc.atoms, kd.symmetry)
        
        # assert self.calc.wfs.nspins == 1
        
    @timer('G0W0')
    def calculate(self, ecuts=None):
        """Starts the G0W0 calculation. Returns a dict with the results with
        the following key/value pairs:

        f: (s, k, n) ndarray
           Occupation numbers
        eps: (s, k, n) ndarray
           Kohn-Sham eigenvalues in eV
        vxc: (s, k, n) ndarray
           Exchange-correlation contributions in eV
        exx: (s, k, n) ndarray
           Exact exchange contributions in eV
        sigma: (s, k, n) ndarray
           Self-energy contributions in eV
        Z: (s, k, n) ndarray
           Renormalization factors
        qp: (s, k, n) ndarray
           Quasi particle energies in eV
        """
        kd = self.calc.wfs.kd

        self.calculate_ks_xc_contribution()
        self.calculate_exact_exchange()

        # Reset calculation
        self.sigma_sin = np.zeros(self.shape)   # self-energies
        self.dsigma_sin = np.zeros(self.shape)  # derivatives of self-energies

        # Get KS eigenvalues and occupation numbers:
        b1, b2 = self.bands
        nibzk = self.calc.wfs.kd.nibzkpts
        for i, k in enumerate(self.kpts):
            for s in range(self.nspins):
                u = s * nibzk + k
                kpt = self.calc.wfs.kpt_u[u]
                self.eps_sin[s, i] = kpt.eps_n[b1:b2]
                self.f_sin[s, i] = kpt.f_n[b1:b2] / kpt.weight

        # My part of the states we want to calculate QP-energies for:
        mykpts = [self.get_k_point(s, K, n1, n2)
                  for s, K, n1, n2 in self.mysKn1n2]
        
        # Loop over q in the IBZ:
        nQ = 0
        for pd0, W0, q_c in self.calculate_screened_potential():
            for kpt1 in mykpts:
                K2 = kd.find_k_plus_q(q_c, [kpt1.K])[0]
                kpt2 = self.get_k_point(kpt1.s, K2, 0, self.nbands, block=True)
                k1 = kd.bz2ibz_k[kpt1.K]
                i = self.kpts.index(k1)
                self.calculate_q(i, kpt1, kpt2, pd0, W0)
            nQ += 1

        self.world.sum(self.sigma_sin)
        self.world.sum(self.dsigma_sin)
        
        self.Z_sin = 1 / (1 - self.dsigma_sin)
        self.qp_sin = self.eps_sin + self.Z_sin * (self.sigma_sin +
                                                   self.exx_sin -
                                                   self.vxc_sin)
        
        results = {'f': self.f_sin,
                   'eps': self.eps_sin * Hartree,
                   'vxc': self.vxc_sin * Hartree,
                   'exx': self.exx_sin * Hartree,
                   'sigma': self.sigma_sin * Hartree,
                   'Z': self.Z_sin,
                   'qp': self.qp_sin * Hartree}
      
        self.print_results(results)

        if self.savepckl:
            pickle.dump(results,
                        paropen(self.filename + '_results.pckl', 'wb'))
        
        return results
        
    def calculate_q(self, i, kpt1, kpt2, pd0, W0):
        """Calculates the contribution to the self-energy and its derivative
        for a given set of k-points, kpt1 and kpt2."""
        wfs = self.calc.wfs
        
        N_c = pd0.gd.N_c
        i_cG = self.sign * np.dot(self.U_cc,
                                  np.unravel_index(pd0.Q_qG[0], N_c))

        q_c = wfs.kd.bzk_kc[kpt2.K] - wfs.kd.bzk_kc[kpt1.K]

        shift0_c = q_c - self.sign * np.dot(self.U_cc, pd0.kd.bzk_kc[0])
        assert np.allclose(shift0_c.round(), shift0_c)
        shift0_c = shift0_c.round().astype(int)

        shift_c = kpt1.shift_c - kpt2.shift_c - shift0_c
        I_G = np.ravel_multi_index(i_cG + shift_c[:, None], N_c, 'wrap')
        
        G_Gv = pd0.get_reciprocal_vectors()
        pos_av = np.dot(self.spos_ac, pd0.gd.cell_cv)
        M_vv = np.dot(pd0.gd.cell_cv.T,
                      np.dot(self.U_cc.T,
                             np.linalg.inv(pd0.gd.cell_cv).T))
        
        Q_aGii = []
        for a, Q_Gii in enumerate(self.Q_aGii):
            x_G = np.exp(1j * np.dot(G_Gv, (pos_av[a] -
                                            np.dot(M_vv, pos_av[a]))))
            U_ii = self.calc.wfs.setups[a].R_sii[self.s]
            Q_Gii = np.dot(np.dot(U_ii, Q_Gii * x_G[:, None, None]),
                           U_ii.T).transpose(1, 0, 2)
            if self.sign == -1:
                Q_Gii = Q_Gii.conj()
            Q_aGii.append(Q_Gii)

        if debug:
            self.check(i_cG, shift0_c, N_c, q_c, Q_aGii)
                
        if self.ppa:
            calculate_sigma = self.calculate_sigma_ppa
        else:
            calculate_sigma = self.calculate_sigma
            
        for n in range(kpt1.n2 - kpt1.n1):
            ut1cc_R = kpt1.ut_nR[n].conj()
            eps1 = kpt1.eps_n[n]
            C1_aGi = [np.dot(Qa_Gii, P1_ni[n].conj())
                      for Qa_Gii, P1_ni in zip(Q_aGii, kpt1.P_ani)]
            n_mG = self.calculate_pair_densities(ut1cc_R, C1_aGi, kpt2,
                                                 pd0, I_G)
            if self.sign == 1:
                n_mG = n_mG.conj()
                                    
            f_m = kpt2.f_n
            deps_m = eps1 - kpt2.eps_n
            sigma, dsigma = calculate_sigma(n_mG, deps_m, f_m, W0)
            nn = kpt1.n1 + n - self.bands[0]
            self.sigma_sin[kpt1.s, i, nn] += sigma
            self.dsigma_sin[kpt1.s, i, nn] += dsigma
            
    def check(self, i_cG, shift0_c, N_c, q_c, Q_aGii):
        I0_G = np.ravel_multi_index(i_cG - shift0_c[:, None], N_c, 'wrap')
        qd1 = KPointDescriptor([q_c])
        pd1 = PWDescriptor(self.ecut, self.calc.wfs.gd, complex, qd1)
        G_I = np.empty(N_c.prod(), int)
        G_I[:] = -1
        I1_G = pd1.Q_qG[0]
        G_I[I1_G] = np.arange(len(I0_G))
        G_G = G_I[I0_G]
        assert len(I0_G) == len(I1_G)
        assert (G_G >= 0).all()
        for a, Q_Gii in enumerate(self.initialize_paw_corrections(pd1)):
            e = abs(Q_aGii[a] - Q_Gii[G_G]).max()
            assert e < 1e-12

    @timer('Sigma')
    def calculate_sigma(self, n_mG, deps_m, f_m, C_swGG):
        """Calculates a contribution to the self-energy and its derivative for
        a given (k, k-q)-pair from its corresponding pair-density and
        energy."""
        o_m = abs(deps_m)
        # Add small number to avoid zeros for degenerate states:
        sgn_m = np.sign(deps_m + 1e-15)
        
        # Pick +i*eta or -i*eta:
        s_m = (1 + sgn_m * np.sign(0.5 - f_m)).astype(int) // 2
        
        beta = (2**0.5 - 1) * self.domega0 / self.omega2
        w_m = (o_m / (self.domega0 + beta * o_m)).astype(int)
        o1_m = self.omega_w[w_m]
        o2_m = self.omega_w[w_m + 1]
        
        x = 1.0 / (self.qd.nbzkpts * 2 * pi * self.vol)
        sigma = 0.0
        dsigma = 0.0
        # Performing frequency integration
        for o, o1, o2, sgn, s, w, n_G in zip(o_m, o1_m, o2_m,
                                             sgn_m, s_m, w_m, n_mG):
            C1_GG = C_swGG[s][w]
            C2_GG = C_swGG[s][w + 1]
            p = x * sgn
            myn_G = n_G[self.Ga:self.Gb]
            sigma1 = p * np.dot(np.dot(myn_G, C1_GG), n_G.conj()).imag
            sigma2 = p * np.dot(np.dot(myn_G, C2_GG), n_G.conj()).imag
            sigma += ((o - o1) * sigma2 + (o2 - o) * sigma1) / (o2 - o1)
            dsigma += sgn * (sigma2 - sigma1) / (o2 - o1)
            
        return sigma, dsigma

    def calculate_screened_potential(self):
        """Calculates the screened potential for each q-point in the 1st BZ.
        Since many q-points are related by symmetry, the actual calculation is
        only done for q-points in the IBZ and the rest are obtained by symmetry
        transformations. Results are returned as a generator to that it is not
        necessary to store a huge matrix for each q-point in the memory."""
        # The decorator $timer('W') doesn't work for generators, do we will
        # have to manually start and stop the timer here:
        self.timer.start('W')
        print('Calculating screened Coulomb potential', file=self.fd)
        if self.truncation is not None:
            print('Using %s truncated Coloumb potential' % self.truncation,
                  file=self.fd)
            
        if self.ppa:
            print('Using Godby-Needs plasmon-pole approximation:',
                  file=self.fd)
            print('    Fitting energy: i*E0, E0 = %.3f Hartee' % self.E0,
                  file=self.fd)

            # use small imaginary frequency to avoid dividing by zero:
            frequencies = [1e-10j, 1j * self.E0 * Hartree]
            
            parameters = {'eta': 0,
                          'hilbert': False,
                          'timeordered': False,
                          'frequencies': frequencies}
        else:
            print('Using full frequency integration:', file=self.fd)
            print('  domega0: {0:g}'.format(self.domega0 * Hartree),
                  file=self.fd)
            print('  omega2: {0:g}'.format(self.omega2 * Hartree),
                  file=self.fd)

            parameters = {'eta': self.eta * Hartree,
                          'hilbert': True,
                          'timeordered': True,
                          'domega0': self.domega0 * Hartree,
                          'omega2': self.omega2 * Hartree}
        
        chi0 = Chi0(self.inputcalc,
                    nbands=self.nbands,
                    ecut=self.ecut * Hartree,
                    intraband=False,
                    real_space_derivatives=False,
                    txt=self.filename + '.w.txt',
                    timer=self.timer,
                    keep_occupied_states=True,
                    nblocks=self.blockcomm.size,
                    **parameters)

        if self.truncation == 'wigner-seitz':
            wstc = WignerSeitzTruncatedCoulomb(
                self.calc.wfs.gd.cell_cv,
                self.calc.wfs.kd.N_c,
                chi0.fd)
        else:
            wstc = None
        
        self.omega_w = chi0.omega_w
        self.omegamax = chi0.omegamax
        
        htp = HilbertTransform(self.omega_w, self.eta, gw=True)
        htm = HilbertTransform(self.omega_w, -self.eta, gw=True)

        # Find maximum size of chi-0 matrices:
        gd = self.calc.wfs.gd
        nGmax = max(count_reciprocal_vectors(self.ecut, gd, q_c)
                    for q_c in self.qd.ibzk_kc)
        nw = len(self.omega_w)
        
        size = self.blockcomm.size
        
        mynGmax = (nGmax + size - 1) // size
        mynw = (nw + size - 1) // size
        
        # Allocate memory in the beginning and use for all q:
        A1_x = np.empty(nw * mynGmax * nGmax, complex)
        A2_x = np.empty(max(mynw * nGmax, nw * mynGmax) * nGmax, complex)
        
        # Need to pause the timer in between iterations
        self.timer.stop('W')
        for iq, q_c in enumerate(self.qd.ibzk_kc):
            self.timer.start('W')
            if self.savew:
                wfilename = self.filename + '.w.q%d.pckl' % iq
                fd = opencew(wfilename)
            if self.savew and fd is None:
                # Read screened potential from file
                with open(wfilename, 'rb') as fd:
                    pd, W = pickle.load(fd)
                # We also need to initialize the PAW corrections
                self.Q_aGii = self.initialize_paw_corrections(pd)
            else:
                # First time calculation
                pd, W = self.calculate_w(chi0, q_c, htp, htm, wstc, A1_x, A2_x)
                if self.savew:
                    pickle.dump((pd, W), fd, pickle.HIGHEST_PROTOCOL)

            self.timer.stop('W')
            # Loop over all k-points in the BZ and find those that are related
            # to the current IBZ k-point by symmetry
            Q1 = self.qd.ibz2bz_k[iq]
            done = set()
            for Q2 in self.qd.bz2bz_ks[Q1]:
                if Q2 >= 0 and Q2 not in done:
                    s = self.qd.sym_k[Q2]
                    self.s = s
                    self.U_cc = self.qd.symmetry.op_scc[s]
                    time_reversal = self.qd.time_reversal_k[Q2]
                    self.sign = 1 - 2 * time_reversal
                    Q_c = self.qd.bzk_kc[Q2]
                    d_c = self.sign * np.dot(self.U_cc, q_c) - Q_c
                    assert np.allclose(d_c.round(), d_c)
                    yield pd, W, Q_c
                    done.add(Q2)
    
    @timer('WW')
    def calculate_w(self, chi0, q_c, htp, htm, wstc, A1_x, A2_x):
        """Calculates the screened potential for a specified q-point."""
        pd, chi0_wGG, chi0_wxvG, chi0_wvv = chi0.calculate(q_c, A_x=A1_x)
        self.Q_aGii = chi0.Q_aGii
        self.Ga = chi0.Ga
        self.Gb = chi0.Gb

        nw = chi0_wGG.shape[0]
        mynw = (nw + self.blockcomm.size - 1) // self.blockcomm.size

        if self.blockcomm.size > 1:
            A1_x = chi0_wGG.ravel()
            chi0_wGG = chi0.redistribute(chi0_wGG, A2_x)
            wa = min(self.blockcomm.rank * mynw, nw)
        else:
            wa = 0

        if self.integrate_gamma != 0:
            if self.integrate_gamma == 2:
                reduced = True
            else:
                reduced = False
            V0, sqrV0 = get_integrated_kernel(pd,
                                              self.calc.wfs.kd.N_c,
                                              truncation=self.truncation,
                                              reduced=reduced,
                                              N=100)
        elif self.integrate_gamma == 0 and np.allclose(q_c, 0):
            bzvol = (2 * np.pi)**3 / self.vol / self.qd.nbzkpts
            Rq0 = (3 * bzvol / (4 * np.pi))**(1. / 3.)
            V0 = 16 * np.pi**2 * Rq0 / bzvol
            sqrV0 = (4 * np.pi)**(1.5) * Rq0**2 / bzvol / 2
        else:
            pass
            
        nG = len(chi0_wGG[0])
        delta_GG = np.eye(nG)

        if self.ppa:
            einv_wGG = []
            
        self.timer.start('Dyson eq.')
        # Calculate W and store it in chi0_wGG ndarray:
        for iw, chi0_GG in enumerate(chi0_wGG):
            if np.allclose(q_c, 0):
                # Generate fine grid in vicinity of gamma
                kd = self.calc.wfs.kd
                N = 4
                N_c = np.array([N, N, N])
                if self.truncation is not None:
                    # Only average periodic directions if trunction is used
                    N_c[np.where(kd.N_c == 1)[0]] = 1
                qf_qc = monkhorst_pack(N_c) / kd.N_c
                qf_qc *= 1.0e-6
                U_scc = kd.symmetry.op_scc
                qf_qc = kd.get_ibz_q_points(qf_qc, U_scc)[0]
                weight_q = kd.q_weights
                qf_qv = 2 * np.pi * np.dot(qf_qc, pd.gd.icell_cv)
                a_q = np.sum(np.dot(chi0_wvv[wa+iw], qf_qv.T) * qf_qv.T, axis=0)
                a0_qG = np.dot(qf_qv, chi0_wxvG[wa+iw, 0])
                a1_qG = np.dot(qf_qv, chi0_wxvG[wa+iw, 1])
                einv_GG = np.zeros((nG, nG), complex)
                # W_GG = np.zeros((nG, nG), complex)
                for iqf in range(len(qf_qv)):
                    chi0_GG[0] = a0_qG[iqf]
                    chi0_GG[:, 0] = a1_qG[iqf]
                    chi0_GG[0, 0] = a_q[iqf]
                    sqrV_G = get_coulomb_kernel(pd,
                                                kd.N_c,
                                                truncation=self.truncation,
                                                wstc=wstc,
                                                q_v=qf_qv[iqf])**0.5
                    e_GG = np.eye(nG) - chi0_GG * sqrV_G * sqrV_G[:,
                                                                  np.newaxis]
                    einv_GG += np.linalg.inv(e_GG) * weight_q[iqf]
                    # einv_GG = np.linalg.inv(e_GG) * weight_q[iqf]
                    # W_GG += (einv_GG * sqrV_G * sqrV_G[:, np.newaxis]
                    #          * weight_q[iqf])
            else:
                sqrV_G = get_coulomb_kernel(pd,
                                            self.calc.wfs.kd.N_c,
                                            truncation=self.truncation,
                                            wstc=wstc)**0.5
                e_GG = (delta_GG - chi0_GG * sqrV_G * sqrV_G[:, np.newaxis])
                einv_GG = np.linalg.inv(e_GG)

            if self.ppa:
                einv_wGG.append(einv_GG - delta_GG)
            else:
                W_GG = chi0_GG
                W_GG[:] = (einv_GG - delta_GG) * sqrV_G * sqrV_G[:, np.newaxis]

                if np.allclose(q_c, 0) or self.integrate_gamma != 0:
                    W_GG[0, 0] = (einv_GG[0, 0] - 1.0) * V0
                    W_GG[0, 1:] = einv_GG[0, 1:] * sqrV_G[1:] * sqrV0
                    W_GG[1:, 0] = einv_GG[1:, 0] * sqrV0 * sqrV_G[1:]
                else:
                    pass
        
        if self.ppa:
            omegat_GG = self.E0 * np.sqrt(einv_wGG[1] /
                                          (einv_wGG[0] - einv_wGG[1]))
            R_GG = -0.5 * omegat_GG * einv_wGG[0]
            W_GG = pi * R_GG * sqrV_G * sqrV_G[:, np.newaxis]
            if np.allclose(q_c, 0) or self.integrate_gamma != 0:
                W_GG[0, 0] = pi * R_GG[0, 0] * V0
                W_GG[0, 1:] = pi * R_GG[0, 1:] * sqrV_G[1:] * sqrV0
                W_GG[1:, 0] = pi * R_GG[1:, 0] * sqrV0 * sqrV_G[1:]

            self.timer.stop('Dyson eq.')
            return pd, [W_GG, omegat_GG]
            
        if self.blockcomm.size > 1:
            Wm_wGG = chi0.redistribute(chi0_wGG, A1_x)
        else:
            Wm_wGG = chi0_wGG
            
        Wp_wGG = A2_x[:Wm_wGG.size].reshape(Wm_wGG.shape)
        Wp_wGG[:] = Wm_wGG

        with self.timer('Hilbert transform'):
            htp(Wp_wGG)
            htm(Wm_wGG)
        self.timer.stop('Dyson eq.')
        
        return pd, [Wp_wGG, Wm_wGG]

    @timer('Kohn-Sham XC-contribution')
    def calculate_ks_xc_contribution(self):
        name = self.filename + '.vxc.npy'
        fd = opencew(name)
        if fd is None:
            print('Reading Kohn-Sham XC contribution from file:', name,
                  file=self.fd)
            with open(name, 'rb') as fd:
                self.vxc_sin = np.load(fd)
            assert self.vxc_sin.shape == self.shape, self.vxc_sin.shape
            return
            
        print('Calculating Kohn-Sham XC contribution', file=self.fd)
        if self.reader is not None:
            self.calc.wfs.read_projections(self.reader)
        vxc_skn = vxc(self.calc, self.calc.hamiltonian.xc) / Hartree
        n1, n2 = self.bands
        self.vxc_sin = vxc_skn[:, self.kpts, n1:n2]
        np.save(fd, self.vxc_sin)
        
    @timer('EXX')
    def calculate_exact_exchange(self):
        name = self.filename + '.exx.npy'
        fd = opencew(name)
        if fd is None:
            print('Reading EXX contribution from file:', name, file=self.fd)
            with open(name, 'rb') as fd:
                self.exx_sin = np.load(fd)
            assert self.exx_sin.shape == self.shape, self.exx_sin.shape
            return
            
        print('Calculating EXX contribution', file=self.fd)
        exx = EXX(self.calc, kpts=self.kpts, bands=self.bands,
                  txt=self.filename + '.exx.txt', timer=self.timer)
        exx.calculate()
        self.exx_sin = exx.get_eigenvalue_contributions() / Hartree
        np.save(fd, self.exx_sin)

    def print_results(self, results):
        description = ['f:     Occupation numbers',
                       'eps:   KS-eigenvalues [eV]',
                       'vxc:   KS vxc [eV]',
                       'exx:   Exact exchange [eV]',
                       'sigma: Self-energies [eV]',
                       'Z:     Renormalization factors',
                       'qp:    QP-energies [eV]']

        print('\nResults:', file=self.fd)
        for line in description:
            print(line, file=self.fd)
            
        b1, b2 = self.bands
        names = [line.split(':', 1)[0] for line in description]
        ibzk_kc = self.calc.wfs.kd.ibzk_kc
        for s in range(self.calc.wfs.nspins):
            for i, ik in enumerate(self.kpts):
                print('\nk-point ' +
                      '{0} ({1}): ({2:.3f}, {3:.3f}, {4:.3f})'.format(
                          i, ik, *ibzk_kc[ik]), file=self.fd)
                print('band' +
                      ''.join('{0:>8}'.format(name) for name in names),
                      file=self.fd)
                for n in range(b2 - b1):
                    print('{0:4}'.format(n + b1) +
                          ''.join('{0:8.3f}'.format(results[name][s, i, n])
                                  for name in names),
                          file=self.fd)

        self.timer.write(self.fd)

    @timer('PPA-Sigma')
    def calculate_sigma_ppa(self, n_mG, deps_m, f_m, W):
        W_GG, omegat_GG = W

        sigma = 0.0
        dsigma = 0.0

        # init variables (is this necessary?)
        nG = n_mG.shape[1]
        deps_GG = np.empty((nG, nG))
        sign_GG = np.empty((nG, nG))
        x1_GG = np.empty((nG, nG))
        x2_GG = np.empty((nG, nG))
        x3_GG = np.empty((nG, nG))
        x4_GG = np.empty((nG, nG))
        x_GG = np.empty((nG, nG))
        dx_GG = np.empty((nG, nG))
        nW_G = np.empty(nG)
        for m in range(np.shape(n_mG)[0]):
            deps_GG = deps_m[m]
            sign_GG = 2 * f_m[m] - 1
            x1_GG = 1 / (deps_GG + omegat_GG - 1j * self.eta)
            x2_GG = 1 / (deps_GG - omegat_GG + 1j * self.eta)
            x3_GG = 1 / (deps_GG + omegat_GG - 1j * self.eta * sign_GG)
            x4_GG = 1 / (deps_GG - omegat_GG - 1j * self.eta * sign_GG)
            x_GG = W_GG * (sign_GG * (x1_GG - x2_GG) + x3_GG + x4_GG)
            dx_GG = W_GG * (sign_GG * (x1_GG**2 - x2_GG**2) +
                            x3_GG**2 + x4_GG**2)
            nW_G = np.dot(n_mG[m], x_GG)
            sigma += np.vdot(n_mG[m], nW_G).real
            nW_G = np.dot(n_mG[m], dx_GG)
            dsigma -= np.vdot(n_mG[m], nW_G).real
        
        x = 1 / (self.qd.nbzkpts * 2 * pi * self.vol)
        return x * sigma, x * dsigma



from __future__ import print_function
import numpy as np
from gpaw.xc.libxc import LibXC
from gpaw.xc.gga import GGA
from gpaw.utilities import compiled_with_libvdwxc
from gpaw.utilities.grid_redistribute import Domains, general_redistribute
from gpaw.utilities.timing import nulltimer
from gpaw.mpi import have_mpi
import _gpaw


def check_grid_descriptor(gd):
    assert gd.parsize_c[1] == 1 and gd.parsize_c[2] == 1
    nxpts_p = gd.n_cp[0][1:] - gd.n_cp[0][:-1]
    nxpts0 = nxpts_p[0]
    for nxpts in nxpts_p[1:-1]:
        assert nxpts == nxpts0
    assert nxpts_p[-1] <= nxpts0


_VDW_NUMERICAL_CODES = {'vdW-DF': 1,
                        'vdW-DF2': 2,
                        'vdW-DF-CX': 3}


def libvdwxc_has_mpi():
    return have_mpi and _gpaw.libvdwxc_has('mpi')


def libvdwxc_has_pfft():
    return have_mpi and _gpaw.libvdwxc_has('pfft')


def get_domains(N_c, parsize_c):
    # We want a distribution like this:
    #   [B, B, ..., B, remainder, 0, 0, ..., 0].
    # with blocksize B chosen as large as possible for better load balance.
    # This function returns the blocksize and the cumulative sum of indices
    # starting with 0.
    blocksize_c = -(-N_c // parsize_c)
    return (np.arange(1 + parsize_c) * blocksize_c).clip(0, N_c)


def get_auto_pfft_grid(size):
    nproc1 = size
    nproc2 = 1
    while nproc1 > nproc2 and nproc1 % 2 == 0:
        nproc1 /= 2
        nproc2 *= 2
    return nproc1, nproc2


class VDWDistribution:
    def __init__(self, gd, parsize_c):
        self.input_gd = gd
        assert np.product(parsize_c) == gd.comm.size
        self.local_input_size_c = gd.n_c
        self.domains_in = Domains(gd.n_cp)
        N_c = gd.get_size_of_global_array(pad=True)
        self.domains_out = Domains([get_domains(N_c[i], parsize_c[i])
                                    for i in range(3)])

        # The auxiliary gd actually is used *only* for the rank/parpos
        # correspondence.  The actual domains it defines are unused!!
        self.aux_gd = gd.new_descriptor(comm=gd.comm, parsize_c=parsize_c)
        parpos_c = self.aux_gd.get_processor_position_from_rank()
        
        self.local_output_size_c = tuple(self.domains_out.get_box(parpos_c)[1])

    def block_zeros(self, shape=(),):
        return np.zeros(shape + self.local_output_size_c)

    def gd2block(self, a_xg, b_xg):
        general_redistribute(self.input_gd.comm,
                             self.domains_in, self.domains_out,
                             self.input_gd.get_processor_position_from_rank,
                             self.aux_gd.get_processor_position_from_rank,
                             a_xg, b_xg, behavior='overwrite')

    def block2gd_add(self, a_xg, b_xg):
        general_redistribute(self.input_gd.comm,
                             self.domains_out, self.domains_in,
                             self.aux_gd.get_processor_position_from_rank,
                             self.input_gd.get_processor_position_from_rank,
                             a_xg, b_xg, behavior='add')


class LibVDWXC(GGA, object):
    def __init__(self, gga_kernel, name, timer=nulltimer, parallel='auto',
                 pfft_grid=None):
        """Initialize LibVDWXC object (further initialization required).

        parallel can be 'auto', '', 'serial', 'mpi', or 'pfft'.

         * 'serial' uses FFTW and only works with serial decompositions.

         * 'mpi' uses FFTW-MPI with communicator of the grid
           descriptor, parallelizing along the x axis.

         * 'pfft' uses PFFT and works with any decomposition,
           parallelizing along two directions for best scalability.

         * 'auto' uses PFFT if available, else FFTW-MPI if available,
           else adhoc if applicable, else serial.

         pfft_grid is the 2D CPU grid used by PFFT and can be a tuple
         (nproc1, nproc2) that multiplies to total communicator size,
         or None.  It is an error to specify pfft_grid unless using
         PFFT.  If left unspecified, a hopefully reasonable automatic
         choice will be made.
         """
        self._vdw = None
        object.__init__(self)
        GGA.__init__(self, gga_kernel)
        self.timer = timer
        self.vdwcoef = 1.0
        self.vdw_functional_name = name

        if parallel == 'auto':
            if have_mpi and libvdwxc_has_mpi():
                parallel = 'mpi'
            else:
                parallel = 'serial'
        self.parallel = parallel
        if parallel != 'pfft' and pfft_grid is not None:
            raise ValueError('pfft_grid specified but pfft not available')
        self.pfft_grid = pfft_grid
        self.distribution = None

        # Check for missing libraries.  I guess ImportError makes most sense?
        if not compiled_with_libvdwxc():
            raise ImportError('libvdwxc not compiled into GPAW')
        if parallel != 'serial' and not have_mpi:
            raise ImportError('MPI not available for libvdwxc-%s '
                              'because GPAW is serial' % parallel)
        if parallel == 'mpi' and not libvdwxc_has_mpi():
            raise ImportError('libvdwxc not compiled with MPI')
        if parallel == 'pfft' and not libvdwxc_has_pfft():
            raise ImportError('libvdwxc not compiled with PFFT')

    @property
    def name(self):
        if self.parallel == 'serial':
            pardesc = 'FFTW serial'
        elif self.parallel == 'mpi':
            if self.distribution is None:
                cores = 'unspecified number of cores'
            else:
                size = self.distribution.input_gd.comm.size
                cores = '%d cores' % size if size != 1 else 'one core'
            pardesc = 'FFTW-MPI with %s' % cores
        else:
            assert self.parallel == 'pfft'
            if self.pfft_grid is None:
                pardesc = 'PFFT with unspecified parallelization'
            else:
                pardesc = 'PFFT with %d x %d CPU grid' % tuple(self.pfft_grid)
        return '%s [libvdwxc/%s]' % (self.vdw_functional_name, pardesc)

    @name.setter
    def name(self, value):
        # Somewhere in the class hierarchy, someone tries to set the name.
        pass

    def get_setup_name(self):
        return 'revPBE'

    def _libvdwxc_init(self, gd):
        """Initialize libvdwxc things in C."""
        self.timer.start('libvdwxc init')
        
        comm = gd.comm

        # TODO Here we could decide FFT padding.
        N_c = gd.get_size_of_global_array(pad=True)

        code = _VDW_NUMERICAL_CODES[self.vdw_functional_name]
        self._vdw = np.empty(1, np.intp)
        _gpaw.libvdwxc_create(self._vdw, code, tuple(N_c),
                              tuple(gd.cell_cv.ravel()))

        cpugrid = [1, 1, 1]

        if self.parallel == 'pfft':
            if self.pfft_grid is None:
                # TODO decide grid intelligently (e.g., based on other comms)
                self.pfft_grid = get_auto_pfft_grid(comm.size)
            nproc1, nproc2 = self.pfft_grid
            assert nproc1 * nproc2 == comm.size
            cpugrid[0] = nproc1
            cpugrid[1] = nproc2
            _gpaw.libvdwxc_init_pfft(self._vdw, comm.get_c_object(),
                                     nproc1, nproc2)
        elif self.parallel == 'mpi':
            cpugrid[0] = comm.size
            _gpaw.libvdwxc_init_mpi(self._vdw, comm.get_c_object())
        else:
            assert self.parallel == 'serial'
            assert comm.size == 1, ('You cannot run in serial with %d cores'
                                    % comm.size)
            _gpaw.libvdwxc_init_serial(self._vdw)

        self.distribution = VDWDistribution(gd, cpugrid)
        self.timer.stop('libvdwxc init')

    def initialize(self, density, hamiltonian, wfs, occupations):
        self.timer.start('initialize')
        GGA.initialize(self, density, hamiltonian, wfs, occupations)
        self._libvdwxc_init(density.finegd)
        self.timer.stop('initialize')

    def calculate(self, gd, n_sg, v_sg, e_g=None):
        """Calculate energy and potential.

        gd may be non-periodic.  To be distinguished from self.gd
        which is always periodic due to priminess of FFT dimensions.
        (To do: proper padded FFTs.)"""
        nspins = len(n_sg)
        assert nspins == 1, 'libvdwxc does not work with multiple spins yet'
        assert gd == self.distribution.input_gd

        if e_g is not None:
            # TODO: handle e_g properly
            raise NotImplementedError('Proper energy density e_g in libvdwxc')
        energy = GGA.calculate(self, gd, n_sg, v_sg, e_g=None)
        return energy

    def calculate_nonlocal(self, n_g, sigma_g, v_g, dedsigma_g):
        for arr in [n_g, sigma_g, v_g, dedsigma_g]:
            assert arr.flags.contiguous
            assert arr.dtype == float
            np.all(arr.shape == self.distribution.local_output_size_c)
        energy = _gpaw.libvdwxc_calculate(self._vdw, n_g, sigma_g,
                                          v_g, dedsigma_g)
        return energy

    def calculate_gga(self, e_g, n_sg, v_sg, sigma_xg, dedsigma_xg):
        assert self._vdw is not None
        self.timer.start('van der Waals')
        n_sg[:] = np.abs(n_sg)  # XXXX What to do about this?
        sigma_xg[:] = np.abs(sigma_xg)
        assert len(n_sg) == 1
        assert len(sigma_xg) == 1
        self.timer.start('semilocal')
        GGA.calculate_gga(self, e_g, n_sg, v_sg, sigma_xg, dedsigma_xg)
        self.timer.stop('semilocal')

        zeros = self.distribution.block_zeros

        nblock_g = zeros()
        sigmablock_g = zeros()
        vblock_g = zeros()
        dedsigmablock_g = zeros()

        self.timer.start('redistribute')
        self.distribution.gd2block(n_sg[0], nblock_g)
        self.distribution.gd2block(sigma_xg[0], sigmablock_g)
        self.timer.stop('redistribute')

        self.timer.start('libvdwxc nonlocal')
        energy_nonlocal = self.calculate_nonlocal(nblock_g, sigmablock_g,
                                                  vblock_g, dedsigmablock_g)
        self.timer.stop('libvdwxc nonlocal')

        for obj in [energy_nonlocal, vblock_g, dedsigmablock_g]:
            obj *= self.vdwcoef

        self.timer.start('redistribute')
        self.distribution.block2gd_add(vblock_g, v_sg[0])
        self.distribution.block2gd_add(dedsigmablock_g, dedsigma_xg[0])
        self.timer.stop('redistribute')

        # XXXXXXXXXXXXXXXX ugly
        e_g[0, 0, 0] += self.vdwcoef * energy_nonlocal / self.gd.dv
        self.timer.stop('van der Waals')

    def estimate_memory(self, mem):
        size = self.distribution.input_gd.bytecount()  # only on average
        mem.subnode('thetas', 20 * size)
        mem.subnode('other', 3 * size)

    def __del__(self):
        if self._vdw is not None:
            _gpaw.libvdwxc_free(self._vdw)


def vdw_df(*args, **kwargs):
    kernel = LibXC('GGA_X_PBE_R+LDA_C_PW')
    return LibVDWXC(gga_kernel=kernel, name='vdW-DF', *args, **kwargs)


def vdw_df2(*args, **kwargs):
    kernel = LibXC('GGA_X_RPW86+LDA_C_PW')
    return LibVDWXC(gga_kernel=kernel, name='vdW-DF2', *args, **kwargs)


def vdw_df_cx(*args, **kwargs):
    kernel = CXGGAKernel()
    return LibVDWXC(gga_kernel=kernel, name='vdW-DF-CX', *args, **kwargs)


# WARNING!  These classes will be deprecated and removed soon.
class VDWDF(LibVDWXC):
    def __init__(self, timer=nulltimer):
        kernel = LibXC('GGA_X_PBE_R+LDA_C_PW')
        LibVDWXC.__init__(self, gga_kernel=kernel, name='vdW-DF',
                          timer=timer)


class VDWDF2(LibVDWXC):
    def __init__(self, timer=nulltimer):
        kernel = LibXC('GGA_X_RPW86+LDA_C_PW')
        LibVDWXC.__init__(self, gga_kernel=kernel, name='vdW-DF2',
                          timer=timer)


class VDWDFCX(LibVDWXC):
    def __init__(self, timer=nulltimer):
        kernel = CXGGAKernel()
        LibVDWXC.__init__(self, gga_kernel=kernel, name='vdW-DF-CX',
                          timer=timer)


class CXGGAKernel:
    def __init__(self, just_kidding=False):
        self.just_kidding = just_kidding
        self.type = 'GGA'
        self.lda_c = LibXC('LDA_C_PW')
        if self.just_kidding:
            self.name = 'rPW86_with_%s' % self.lda_c.name
        else:
            self.name = 'CX'

    def calculate(self, e_g, n_sg, v_sg, sigma_xg, dedsigma_xg):
        e_g[:] = 0.0
        dedsigma_xg[:] = 0.0

        self.lda_c.calculate(e_g, n_sg, v_sg, sigma_xg, dedsigma_xg)

        for arr in [n_sg, v_sg, sigma_xg, dedsigma_xg]:
            assert len(arr) == 1
        self._exchange(n_sg[0], sigma_xg[0], e_g, v_sg[0], dedsigma_xg[0])

    def _exchange(self, rho, grho, sx, v1x, v2x):
        """Calculate cx local exchange.

        Note that this *adds* to the energy density sx so that it can
        be called after LDA correlation part without ruining anything.
        Also it adds to v1x and v2x as is normal in GPAW."""
        tol = 1e-20
        rho[rho < tol] = tol
        grho[grho < tol] = tol
        alp = 0.021789
        beta = 1.15
        a = 1.851
        b = 17.33
        c = 0.163
        mu_LM = 0.09434
        s_prefactor = 6.18733545256027
        Ax = -0.738558766382022 # = -3./4. * (3./pi)**(1./3)
        four_thirds = 4. / 3.

        grad_rho = np.sqrt(grho)

        # eventually we need s to power 12.  Clip to avoid overflow
        # (We have individual tolerances on both rho and grho, but
        # they are not sufficient to guarantee this)
        s_1 = (grad_rho / (s_prefactor * rho**four_thirds)).clip(0.0, 1e20)
        s_2 = s_1 * s_1
        s_3 = s_2 * s_1
        s_4 = s_3 * s_1
        s_5 = s_4 * s_1
        s_6 = s_5 * s_1

        fs_rPW86 = (1.0 + a * s_2 + b * s_4 + c * s_6)**(1./15.)

        if self.just_kidding:
            fs = fs_rPW86
        else:
            fs = (1.0 + mu_LM * s_2) / (1.0 + alp * s_6) \
                + alp * s_6 / (beta + alp * s_6) * fs_rPW86

        # the energy density for the exchange.
        sx[:] += Ax * rho**four_thirds * fs

        df_rPW86_ds = (1. / (15. * fs_rPW86**14.0)) * \
            (2 * a * s_1 + 4 * b * s_3 + 6 * c * s_5)

        if self.just_kidding:
            df_ds = df_rPW86_ds # XXXXXXXXXXXXXXXXXXXX
        else:
            df_ds = 1. / (1. + alp * s_6)**2 \
                * (2.0 * mu_LM * s_1 * (1. + alp * s_6)
                   - 6.0 * alp * s_5 * (1. + mu_LM * s_2)) \
                + alp * s_6 / (beta + alp * s_6) * df_rPW86_ds \
                + 6.0 * alp * s_5 * fs_rPW86 / (beta + alp * s_6) \
                * (1. - alp * s_6 / (beta + alp * s_6))

        # de/dn.  This is the partial derivative of sx wrt. n, for s constant
        v1x[:] += Ax * four_thirds * (rho**(1. / 3.) * fs
                                      - grad_rho / (s_prefactor * rho) * df_ds)
        # de/d(nabla n).  The other partial derivative
        v2x[:] += 0.5 * Ax * df_ds / (s_prefactor * grad_rho)
        # (We may or may not understand what that grad_rho is doing here.)


def test_derivatives():
    gen = np.random.RandomState(1)
    shape = (1, 20, 20, 20)
    ngpts = np.product(shape)
    n_sg = gen.rand(*shape)
    sigma_xg = np.zeros(shape)
    sigma_xg[:] = gen.rand(*shape)

    qe_kernel = CXGGAKernel(just_kidding=True)
    libxc_kernel = LibXC('GGA_X_RPW86+LDA_C_PW')

    cx_kernel = CXGGAKernel(just_kidding=False)

    def check(kernel, n_sg, sigma_xg):
        e_g = np.zeros(shape[1:])
        dedn_sg = np.zeros(shape)
        dedsigma_xg = np.zeros(shape)
        kernel.calculate(e_g, n_sg, dedn_sg, sigma_xg, dedsigma_xg)
        return e_g, dedn_sg, dedsigma_xg

    def check_and_write(kernel):
        n1_sg = n_sg.copy()
        e_g, dedn_sg, dedsigma_xg = check(kernel, n_sg, sigma_xg)
        dedn = dedn_sg[0, 0, 0, 0]
        dedsigma = dedsigma_xg[0, 0, 0, 0]

        dn = 1e-6
        n1_sg = n_sg.copy()
        n1_sg[0, 0, 0, 0] -= dn / 2.
        e1_g, _, _ = check(kernel, n1_sg, sigma_xg)

        n1_sg[0, 0, 0, 0] += dn
        e2_g, _, _ = check(kernel, n1_sg, sigma_xg)

        dedn_fd = (e2_g[0, 0, 0] - e1_g[0, 0, 0]) / dn
        dedn_err = abs(dedn - dedn_fd)

        print('e', e_g.sum() / ngpts)
        print('dedn', dedn, 'fd', dedn_fd, 'err %e' % dedn_err)

        sigma1_xg = sigma_xg.copy()
        sigma1_xg[0, 0, 0, 0] -= dn / 2.
        e1s_g, _, _ = check(kernel, n_sg, sigma1_xg)

        sigma1_xg[0, 0, 0, 0] += dn
        e2s_g, _, _ = check(kernel, n_sg, sigma1_xg)

        dedsigma_fd = (e2s_g[0, 0, 0] - e1s_g[0, 0, 0]) / dn
        dedsigma_err = dedsigma - dedsigma_fd

        print('dedsigma', dedsigma, 'fd', dedsigma_fd, 'err %e' % dedsigma_err)
        return e_g, dedn_sg, dedsigma_xg

    print('pw86r libxc')
    e_lxc_g, dedn_lxc_g, dedsigma_lxc_g = check_and_write(libxc_kernel)
    print()
    print('pw86r ours')
    e_qe_g, dedn_qe_g, dedsigma_qe_g = check_and_write(qe_kernel)
    print()
    print('cx')
    check_and_write(cx_kernel)

    print()
    print('lxc vs qe discrepancies')
    print('=======================')
    e_err = np.abs(e_lxc_g - e_qe_g).max()
    print('e', e_err)
    dedn_err = np.abs(dedn_qe_g - dedn_lxc_g).max()
    dedsigma_err = np.abs(dedsigma_lxc_g - dedsigma_qe_g).max()
    print('dedn', dedn_err)
    print('dedsigma', dedsigma_err)


def test_selfconsistent():
    from gpaw import GPAW
    from ase.structure import molecule
    from gpaw.xc.gga import GGA

    system = molecule('H2O')
    system.center(vacuum=3.)

    def test(xc):
        calc = GPAW(mode='lcao',
                    xc=xc,
                    setups='sg15',
                    txt='gpaw.%s.txt' % str(xc)#.kernel.name
                    )
        system.set_calculator(calc)
        return system.get_potential_energy()

    libxc_results = {}

    for name in ['GGA_X_PBE_R+LDA_C_PW', 'GGA_X_RPW86+LDA_C_PW']:
        xc = GGA(LibXC(name))
        e = test(xc)
        libxc_results[name] = e


    cx_gga_results = {}
    cx_gga_results['rpw86'] = test(GGA(CXGGAKernel(just_kidding=True)))
    cx_gga_results['lv_rpw86'] = test(GGA(CXGGAKernel(just_kidding=False)))
    
    vdw_results = {}
    vdw_coef0_results = {}

    for vdw in [VDWDF(), VDWDF2(), VDWDFCX()]:
        vdw.vdwcoef = 0.0
        vdw_coef0_results[vdw.__class__.__name__] = test(vdw)
        vdw.vdwcoef = 1.0 # Leave nicest text file by running real calc last
        vdw_results[vdw.__class__.__name__] = test(vdw)
    
    from gpaw.mpi import world
    # These tests basically verify that the LDA/GGA parts of vdwdf
    # work correctly.
    if world.rank == 0:
        print('Now comparing...')
        err1 = cx_gga_results['rpw86'] - libxc_results['GGA_X_RPW86+LDA_C_PW']
        print('Our rpw86 must be identical to that of libxc. Err=%e' % err1)
        print('RPW86 interpolated with Langreth-Vosko stuff differs by %f'
              % (cx_gga_results['lv_rpw86'] - cx_gga_results['rpw86']))
        print('Each vdwdf with vdwcoef zero must yield same result as gga'
              'kernel')
        err_df1 = vdw_coef0_results['VDWDF'] - libxc_results['GGA_X_PBE_R+'
                                                             'LDA_C_PW']
        print('  df1 err=%e' % err_df1)
        err_df2 = vdw_coef0_results['VDWDF2'] - libxc_results['GGA_X_RPW86+'
                                                              'LDA_C_PW']
        print('  df2 err=%e' % err_df2)
        err_cx = vdw_coef0_results['VDWDFCX'] - cx_gga_results['lv_rpw86']
        print('   cx err=%e' % err_cx)


if __name__ == '__main__':
    test_derivatives()
    #test_selfconsistent()



from ase import Atoms
from gpaw import GPAW, FermiDirac, Mixer, __version__
from gpaw.lcaotddft import LCAOTDDFT
from gpaw.mpi import world
from gpaw.tddft import TDDFT as GRIDTDDFT
from gpaw.test import equal

ref_values = [[-0.0153712, -0.0153712, -0.0153712],
              [ 0.0037263,  0.0037263,  0.0037263],
              [ 0.0118144,  0.0118144,  0.0118144],
              [-0.0902301, -0.0902301, -0.0902301],
              [-0.0835503, -0.0835503, -0.08355028]]
calcs = []

# Cross-check fd and lcao with PBE and GLLBSC.
# Also test GLLBSC-ground state + ALDA-response in both modes.
for (mode, TDDFT) in [('lcao', LCAOTDDFT),
                      ('fd',   GRIDTDDFT)]:
    for (xc, fxc) in [('PBE', 'PBE'),
                      ('GLLBSC', 'GLLBSC'),
                      ('GLLBSC', 'LDA')]:

        if xc=='PBE' and mode=='fd': # There are other checks for this combination
            continue

        tag = 'np%i_ver%s_%s_%s+%s' % (world.size, __version__, mode, xc, fxc)

        # silane
        atoms = Atoms('SiH4',[( 0.0000,  0.0000,  0.0000),
                              ( 0.8544,  0.8544,  0.8544),
                              (-0.8544, -0.8544,  0.8544),
                              ( 0.8544, -0.8544, -0.8544),
                              (-0.8544,  0.8544, -0.8544)])
        atoms.center(vacuum=4.0)

        # 1) ground state
        calcs.append(GPAW(nbands  = 7,
                          mixer   = Mixer(0.20, 5, weight=50.0),
                          convergence = {'eigenstates': 1e-6,
                                         'density':     1e-6,
                                         'energy':      1e-6},
                          h       = 0.40,
                          mode    = mode,
                          xc      = xc,
                          basis   = 'dzp',
                          dtype   = complex))

        atoms.set_calculator(calcs[-1])
        energy = atoms.get_potential_energy()
        calcs[-1].write('gs.gpw', mode='all')

        # 2) restart ground state
        calcs.append(GPAW('gs.gpw'))
        calcs[-1].set(occupations=FermiDirac(0.05))
        calcs[-1].get_potential_energy()
        calcs[-1].write('gs.gpw', mode='all')
        
        # 3) time propagation
        calcs.append(TDDFT('gs.gpw'))
        if fxc != xc:
            calcs[-1].linearize_to_xc(fxc)

        calcs[-1].propagate(10.0, 20, 'dm.%s.dat' % tag)
        # dipole moment must remain zero
        calcs[-1].attach(equal, 1, calcs[-1].density.finegd.calculate_dipole_moment(calcs[-1].density.rhot_g), 0.0, 1.0e-6)
        calcs[-1].write('td.gpw', mode='all')
        
        # 4) restart time propagation + apply kick
        if xc==fxc: # TODO: restart when linearize_to_xc was applied
            calcs.append(TDDFT('td.gpw'))

        calcs[-1].absorption_kick([0.01, 0.01, 0.01])
        calcs[-1].propagate(10.0, 30, 'dm.%s.dat' % tag)

        dens = calcs[-1].density
        equal(dens.finegd.calculate_dipole_moment(dens.rhot_g),
              ref_values.pop(0), 1.0e-5,
              msg="Failed with %s/%s+%s: " % (mode, xc, fxc))



from __future__ import print_function
from ase import *
from ase.dft.kpoints import monkhorst_pack
from ase.lattice import bulk
from gpaw import *
from gpaw.test import equal
import numpy as np

a0 = 5.43
cell = bulk('Si', 'fcc', a=a0).get_cell()
Si = Atoms('Si2', cell=cell, pbc=True,
           scaled_positions=((0,0,0), (0.25,0.25,0.25)))

kpts = monkhorst_pack((2,2,2))
kpts += np.array([1/4., 1/4., 1/4.])

calc = GPAW(h=0.18,
            kpts=kpts,
#            parallel={'domain':1}, idiotproof=False,
            occupations=FermiDirac(0.001))
Si.set_calculator(calc)
E = Si.get_potential_energy()

#from gpaw.xc.hybridq import HybridXC
#exx = HybridXC('EXX')
#E_q = E + calc.get_xc_difference(exx)

from gpaw.xc.hybridk import HybridXC
exx = HybridXC('EXX', acdf=True, etotflag=True)
E_k1 = E + calc.get_xc_difference(exx)

from gpaw.xc.hybridk import HybridXC
exx = HybridXC('EXX', acdf=False, etotflag=True)
E_k2 = E + calc.get_xc_difference(exx)

#print 'Hartree-Fock ACDF method (hybridq.py)   :', E_q
print('Hartree-Fock ACDF method (hybridk.py)   :', E_k1)
print('Hartree-Fock Standard method            :', E_k2)

equal(E_k2, E_k1, 0.001)
#equal(E_q, E_k1, 0.001)
equal(E_k1, -27.71, 0.01)



from ase.calculators.jacapo.pw91_psp import defaultpseudopotentials as psp

from ase.tasks.calcfactory import calculator_factory

from ase.test.tasks.dcdft import DeltaCodesDFTTask as Task

xc = 'PBE'

fit = (5, 0.02)

w = 0.06

pw = 700
dw = 1000
kd = 8.0

tag = 'dcdft_%s_jacapo' % xc.lower()

calcfactory = calculator_factory('jacapo',
                                 pw=pw,
                                 dw=dw,
                                 xc='PBE',
                                 #convergence={
                                 #'energy':0.00001,
                                 #'density':0.000001, # 0.0001
                                 #'occupation':0.001, # 0.001
                                 #'maxsteps':None,
                                 #'maxtime':None
                                 #},
                                 ft=w,
                                 symmetry=True,
                                 spinpol=True, # must set explicitly
                                 calculate_stress=False,
                                 deletenc=True,  # start fresh every time
                                 kptdensity=kd,
                                 )

taskopts = {}

task = Task(
    calcfactory=calcfactory,
    tag=tag,
    fit=fit,
    use_lock_files=True,
    **taskopts)

if __name__ == '__main__':
    keys = set(psp.keys()).intersection(set(task.collection.names))
    for m in ['Be', 'Cr', 'Kr', 'Xe']:
        keys.remove(m)  # EOS fails: maybe need higher ecut?
    for m in ['Zn', 'Zr']:
        keys.remove(m)  # do not converge
    # run just only system to check if scripts work
    keys = ['Si']
    task.run(keys)



from __future__ import print_function
import sys

from ase.dft.dos import DOS

from gpaw import GPAW


def dos(filename, plot=False, output='dos.csv', width=0.1):
    """Calculate density of states.
    
    filename: str
        Name of restart-file.
    plot: bool
        Show a plot.
    output: str
        Name of CSV output file.
    width: float
        Width of Gaussians.
    """
    calc = GPAW(filename, txt=None)
    dos = DOS(calc, width)
    D = [dos.get_dos(spin) for spin in range(calc.get_number_of_spins())]
    if output:
        fd = sys.stdout if output == '-' else open(output, 'w')
        for x in zip(dos.energies, *D):
            print(*x, sep=', ', file=fd)
        if output != '-':
            fd.close()
    if plot:
        import matplotlib.pyplot as plt
        for y in D:
            plt.plot(dos.energies, y)
        plt.show()



from ase import Atoms
from gpaw import GPAW
from gpaw.tddft import TDDFT
from gpaw.inducedfield.inducedfield_tddft import TDDFTInducedField
import numpy as np

# Na2 cluster
atoms = Atoms(symbols='Na2',
              positions=[(0, 0, 0), (3.0, 0, 0)],
              pbc=False)
atoms.center(vacuum=3.0)

# Standard ground state calculation
calc = GPAW(nbands=2, h=0.6, setups={'Na': '1'})
atoms.set_calculator(calc)
energy = atoms.get_potential_energy()
calc.write('na2_gs.gpw', mode='all')

# Standard time-propagation initialization
time_step = 10.0
iterations = 60
kick_strength = [1.0e-3, 1.0e-3, 0.0]
td_calc = TDDFT('na2_gs.gpw')
td_calc.absorption_kick(kick_strength=kick_strength)

# Create and attach InducedField object
frequencies = [1.0, 2.08]     # Frequencies of interest in eV
folding = 'Gauss'             # Folding function
width = 0.1                   # Line width for folding in eV
ind = TDDFTInducedField(paw=td_calc,
                        frequencies=frequencies,
                        folding=folding,
                        width=width,
                        restart_file='na2_td.ind')

# Propagate as usual
td_calc.propagate(time_step, iterations // 2, 'na2_td_dm.dat', 'na2_td.gpw')

# Save TDDFT and InducedField objects
td_calc.write('na2_td.gpw', mode='all')
ind.write('na2_td.ind')

ind.paw = None

# Restart and continue
td_calc = TDDFT('na2_td.gpw')

# Load and attach InducedField object
ind = TDDFTInducedField(filename='na2_td.ind',
                        paw=td_calc,
                        restart_file='na2_td.ind')
    
# Continue propagation as usual
td_calc.propagate(time_step, iterations // 2, 'na2_td_dm.dat', 'na2_td.gpw')

# Calculate induced electric field
ind.calculate_induced_field(gridrefinement=2, from_density='comp',
                            poisson_eps=2e-10)

# Test
from gpaw.test import equal
tol = 0.0001
val1 = ind.fieldgd.integrate(ind.Ffe_wg[0])
val2 = ind.fieldgd.integrate(np.abs(ind.Fef_wvg[0][0]))
val3 = ind.fieldgd.integrate(np.abs(ind.Fef_wvg[0][1]))
val4 = ind.fieldgd.integrate(np.abs(ind.Fef_wvg[0][2]))
val5 = ind.fieldgd.integrate(ind.Ffe_wg[1])
val6 = ind.fieldgd.integrate(np.abs(ind.Fef_wvg[1][0]))
val7 = ind.fieldgd.integrate(np.abs(ind.Fef_wvg[1][1]))
val8 = ind.fieldgd.integrate(np.abs(ind.Fef_wvg[1][2]))

equal(val1, 3307.77279745, tol)
equal(val2, 2.24614158834, tol)
equal(val3, 2.20381741663, tol)
equal(val4, 1.69244172329, tol)
equal(val5, 3305.78925228, tol)
equal(val6, 2.09432584636, tol)
equal(val7, 2.09849354306, tol)
equal(val8, 1.59727445257, tol)
    
ind.paw = None



import numpy as np

from gpaw.utilities.blas import gemm
from gpaw.utilities import unpack
from gpaw.utilities.partition import EvenPartitioning


def get_atomic_correction(name):
    cls = dict(dense=DenseAtomicCorrection,
               distributed=DistributedAtomicCorrection,
               scipy=ScipyAtomicCorrection)[name]
    return cls()


class BaseAtomicCorrection:
    name = 'base'
    description = 'base class for atomic corrections with LCAO'

    def __init__(self):
        self.nops = 0

    def redistribute(self, wfs, dX_asp, type='asp', op='forth'):
        assert hasattr(dX_asp, 'redistribute'), dX_asp
        assert op in ['back', 'forth']
        return dX_asp

    def calculate_hamiltonian(self, wfs, kpt, dH_asp, H_MM, yy):
        avalues = self.get_a_values()

        dH_aii = dH_asp.partition.arraydict([setup.dO_ii.shape
                                             for setup in wfs.setups],
                                            dtype=wfs.dtype)
        
        for a in avalues:
            dH_aii[a][:] = yy * unpack(dH_asp[a][kpt.s])
        self.calculate(wfs, kpt.q, dH_aii, H_MM)

    def add_overlap_correction(self, wfs, S_qMM):
        avalues = self.get_a_values()
        dS_aii = [wfs.setups[a].dO_ii for a in avalues]
        dS_aii = dict(zip(avalues, dS_aii))  # XXX get rid of dict

        for a in dS_aii:
            dS_aii[a] = np.asarray(dS_aii[a], wfs.dtype)

        for q, S_MM in enumerate(S_qMM):
            self.calculate(wfs, q, dS_aii, S_MM)

    def gobble_data(self, wfs):
        pass  # Prepare internal data structures for calculate().

    def calculate(self, wfs, q, dX_aii, X_MM):
        raise NotImplementedError

    def get_a_values(self):
        raise NotImplementedError

    def implements_distributed_projections(self):
        return False

    def calculate_projections(self, wfs, kpt):
        for a, P_ni in kpt.P_ani.items():
            # ATLAS can't handle uninitialized output array:
            P_ni.fill(117)
            gemm(1.0, wfs.P_aqMi[a][kpt.q], kpt.C_nM, 0.0, P_ni, 'n')


class DenseAtomicCorrection(BaseAtomicCorrection):
    name = 'dense'
    description = 'dense with blas'

    def gobble_data(self, wfs):
        self.initialize(wfs.P_aqMi, wfs.ksl.Mstart, wfs.ksl.Mstop)
        self.orig_partition = wfs.atom_partition  # XXXXXXXXXXXXXXXXXXXXX

    def initialize(self, P_aqMi, Mstart, Mstop):
        self.P_aqMi = P_aqMi
        self.Mstart = Mstart
        self.Mstop = Mstop

    def get_a_values(self):
        return self.P_aqMi.keys()

    def calculate(self, wfs, q, dX_aii, X_MM):
        dtype = X_MM.dtype
        nops = 0
        for a, dX_ii in dX_aii.items():
            P_Mi = self.P_aqMi[a][q]
            assert dtype == P_Mi.dtype
            dXP_iM = np.zeros((dX_ii.shape[1], P_Mi.shape[0]), dtype)
            # (ATLAS can't handle uninitialized output array)
            gemm(1.0, P_Mi, dX_ii, 0.0, dXP_iM, 'c')
            nops += dXP_iM.size * dX_ii.shape[0]
            gemm(1.0, dXP_iM, P_Mi[self.Mstart:self.Mstop], 1.0, X_MM)
            nops += X_MM.size * dXP_iM.shape[0]
        self.nops = nops


class DistributedAtomicCorrection(BaseAtomicCorrection):
    name = 'distributed'
    description = 'distributed and block-sparse'

    def gobble_data(self, wfs):
        self.orig_partition = wfs.atom_partition
        evenpart = EvenPartitioning(self.orig_partition.comm,
                                    self.orig_partition.natoms)
        self.even_partition = evenpart.as_atom_partition()
    
    def get_a_values(self):
        return self.orig_partition.my_indices  # XXXXXXXXXX
        #return self.even_partition.my_indices

    def redistribute(self, wfs, dX_asp, type='asp', op='forth'):
        if type not in ['asp', 'aii']:
            raise ValueError('Unknown matrix type "%s"' % type)

        # just distributed over gd comm.  It's not the most aggressive
        # we can manage but we want to make band parallelization
        # a bit easier and it won't really be a problem.  I guess
        #
        # Also: This call is blocking, but we could easily do a
        # non-blocking version as we only need this stuff after
        # doing tons of real-space work.

        # XXXXXXXXXXXXXXXXXX
        if 1:
            return dX_asp.deepcopy()

        dX_asp = dX_asp.deepcopy()
        if op == 'forth':
            even = self.orig_partition.as_even_partition()
            dX_asp.redistribute(even)
        else:
            assert op == 'back'
            dX_asp.redistribute(self.orig_partition)
        return dX_asp

    def calculate(self, wfs, q, dX_aii, X_MM):
        # XXX reduce according to kpt.q
        dtype = wfs.dtype
        M_a = wfs.setups.M_a
        nM_a = np.array([setup.nao for setup in wfs.setups])
        Mstart = wfs.ksl.Mstart
        Mstop = wfs.ksl.Mstop

        # Now calculate basis-projector-basis overlap: a1 -> a3 -> a2
        #
        # specifically:
        #   < phi[a1] | p[a3] > * dX[a3] * < p[a3] | phi[a2] >
        #
        # This matrix multiplication is semi-sparse.  It works by blocks
        # of atoms, looping only over pairs that do have nonzero
        # overlaps.  But it might be even nicer with scipy sparse.
        # This we will have to check at some point.
        #
        # The projection arrays P_aaim are distributed over the grid,
        # whereas the X_MM is distributed over the band comm.
        # One could choose a set of a3 to optimize the load balance.
        # Right now the load balance will be "random" and probably
        # not very good.
        #innerloops = 0

        setups = wfs.setups

        outer = 0
        inner = 0
        nops = 0

        nao_a = [setup.nao for setup in setups]

        for (a3, a1), P1_qim in wfs.P_aaqim.items():
            P1_im = P1_qim[q]
            a1M1 = M_a[a1]
            nM1 = nM_a[a1]
            a1M2 = a1M1 + nM1

            if a1M1 > Mstop or a1M2 < Mstart:
                continue

            stickout1 = max(0, Mstart - a1M1)
            stickout2 = max(0, a1M2 - Mstop)
            P1_mi = np.conj(P1_im.T[stickout1:nM1 - stickout2])
            dX_ii = dX_aii[a3]
            X_mM = X_MM[a1M1 + stickout1 - Mstart:a1M2 - stickout2 - Mstart]

            P1dX_mi = np.dot(P1_mi, dX_ii)
            nops += P1dX_mi.size * dX_ii.shape[0]
            outer += 1

            assert len(wfs.P_neighbors_a[a3]) > 0
            if 0:
                for a2 in wfs.P_neighbors_a[a3]:
                    inner += 1
                    # We can use symmetry somehow.  Since the entire matrix
                    # is symmetrized after the X_MM is constructed,
                    # at least in the non-Gamma-point case, we should do
                    # so conditionally somehow.  Right now let's stay out
                    # of trouble.

                    # Humm.  The following works with gamma point
                    # but not with kpts.  XXX take a look at this.
                    # also, it doesn't work with a2 < a1 for some reason.
                    #if a2 > a1:
                    #    continue
                    a2M1 = M_a[a2]
                    a2M2 = a2M1 + wfs.setups[a2].nao
                    P2_im = wfs.P_aaqim[(a3, a2)][q]
                    P1dXP2_mm = np.dot(P1dX_mi, P2_im)
                    X_mM[:, a2M1:a2M2] += P1dXP2_mm
                    #innerloops += 1
                    #if wfs.world.rank == 0:
                    #    print 'y', y
                    #if a1 != a2:
                    #    X_MM[a2M1:a2M2, a1M1:a1M2] += P1dXP2_mm.T.conj()

            a2_a = wfs.P_neighbors_a[a3]

            a2nao = sum(nao_a[a2] for a2 in a2_a)

            P2_iam = np.empty((P1_mi.shape[1], a2nao), dtype)
            m2 = 0
            for a2 in a2_a:
                P2_im = wfs.P_aaqim[(a3, a2)][q]
                nao = nao_a[a2]
                P2_iam[:, m2:m2 + nao] = P2_im
                m2 += nao

            P1dXP2_mam = np.zeros((P1dX_mi.shape[0],
                                   P2_iam.shape[1]), dtype)
            gemm(1.0, P2_iam, P1dX_mi, 0.0, P1dXP2_mam)
            nops += P1dXP2_mam.size * P2_iam.shape[0]

            m2 = 0
            for a2 in a2_a:
                nao = nao_a[a2]
                X_mM[:, M_a[a2]:M_a[a2] + setups[a2].nao] += \
                    P1dXP2_mam[:, m2:m2 + nao]
                m2 += nao

        self.nops = nops
        self.inner = inner
        self.outer = outer

        
class ScipyAtomicCorrection(DistributedAtomicCorrection):
    name = 'scipy'
    description = 'distributed and sparse using scipy'

    def __init__(self, tolerance=1e-12):
        DistributedAtomicCorrection.__init__(self)
        from scipy import sparse
        self.sparse = sparse
        self.tolerance = tolerance

    def gobble_data(self, wfs):
        DistributedAtomicCorrection.gobble_data(self, wfs)
        nq = len(wfs.kd.ibzk_qc)

        I_a = [0]
        I_a.extend(np.cumsum([setup.ni for setup in wfs.setups[:-1]]))
        I = I_a[-1] + wfs.setups[-1].ni
        self.I = I
        self.I_a = I_a

        M_a = wfs.setups.M_a
        M = M_a[-1] + wfs.setups[-1].nao
        nao_a = [setup.nao for setup in wfs.setups]
        ni_a = [setup.ni for setup in wfs.setups]

        Psparse_qIM = [self.sparse.lil_matrix((I, M), dtype=wfs.dtype)
                       for _ in range(nq)]

        for (a3, a1), P_qim in wfs.P_aaqim.items():
            if self.tolerance > 0:
                P_qim = P_qim.copy()
                approximately_zero = np.abs(P_qim) < self.tolerance
                P_qim[approximately_zero] = 0.0
            for q in range(nq):
                Psparse_qIM[q][I_a[a3]:I_a[a3] + ni_a[a3],
                               M_a[a1]:M_a[a1] + nao_a[a1]] = P_qim[q]
        
        self.Psparse_qIM = [x.tocsr() for x in Psparse_qIM]

    def calculate(self, wfs, q, dX_aii, X_MM):
        Mstart = wfs.ksl.Mstart
        Mstop = wfs.ksl.Mstop

        Psparse_IM = self.Psparse_qIM[q]

        dXsparse_II = self.sparse.lil_matrix((self.I, self.I), dtype=wfs.dtype)
        avalues = sorted(dX_aii.keys())
        for a in avalues:
            I1 = self.I_a[a]
            I2 = I1 + wfs.setups[a].ni
            dXsparse_II[I1:I2, I1:I2] = dX_aii[a]
        dXsparse_II = dXsparse_II.tocsr()
        
        Psparse_MI = Psparse_IM[:, Mstart:Mstop].transpose().conjugate()
        Xsparse_MM = Psparse_MI.dot(dXsparse_II.dot(Psparse_IM))
        X_MM[:, :] += Xsparse_MM.todense()

    def calculate_projections(self, wfs, kpt):
        if self.implements_distributed_projections():
            P_In = self.Psparse_qIM[kpt.q].dot(kpt.C_nM.T)
            #for a in self.even_partition.my_indices: # XXXXXXX
            for a in self.orig_partition.my_indices:
                I1 = self.I_a[a]
                I2 = I1 + wfs.setups[a].ni
                kpt.P_ani[a][:, :] = P_In[I1:I2, :].T.conj()
        else:
            DistributedAtomicCorrection.calculate_projections(self, wfs, kpt)

    def implements_distributed_projections(self):
        return True



from __future__ import print_function
from functools import partial

import numpy as np
from gpaw.utilities.grid_redistribute import general_redistribute
from gpaw.utilities.partition import AtomPartition, AtomicMatrixDistributor


class Grid2Grid:
    def __init__(self, comm, broadcast_comm, gd, big_gd, enabled=True):
        self.comm = comm
        self.broadcast_comm = broadcast_comm
        self.gd = gd
        self.big_gd = big_gd
        self.enabled = enabled
        
        if big_gd.comm.rank == 0:
            big_ranks = gd.comm.translate_ranks(big_gd.comm,
                                                np.arange(gd.comm.size))
        else:
            big_ranks = np.empty(gd.comm.size, dtype=int)
        big_gd.comm.broadcast(big_ranks, 0)
        
        bigrank2rank = dict(zip(big_ranks, np.arange(gd.comm.size)))
        def rank2parpos1(rank):
            if rank in bigrank2rank:
                return gd.get_processor_position_from_rank(bigrank2rank[rank])
            else:
                return None

        rank2parpos2 = big_gd.get_processor_position_from_rank

        self._distribute = partial(general_redistribute, big_gd.comm,
                                   gd.n_cp, big_gd.n_cp,
                                   rank2parpos1, rank2parpos2)
        self._collect = partial(general_redistribute, big_gd.comm,
                                big_gd.n_cp, gd.n_cp,
                                rank2parpos2, rank2parpos1)
    
    def distribute(self, src_xg, dst_xg=None):
        if dst_xg is None:
            dst_xg = self.big_gd.empty(src_xg.shape[:-3], dtype=src_xg.dtype)
        self._distribute(src_xg, dst_xg)
        return dst_xg
    
    def collect(self, src_xg, dst_xg=None):
        if dst_xg is None:
            dst_xg = self.gd.empty(src_xg.shape[:-3], src_xg.dtype)
        self._collect(src_xg, dst_xg)
        self.broadcast_comm.broadcast(dst_xg, 0)
        return dst_xg

    # Strangely enough the purpose of this is to appease AtomPAW
    def new(self, gd, big_gd):
        return Grid2Grid(self.comm, self.broadcast_comm, gd, big_gd,
                         self.enabled)

    def get_matrix_distributor(self, atom_partition, spos_ac=None):
        if spos_ac is None:
            rank_a = np.zeros(self.big_gd.comm.size, dtype=int)
        else:
            rank_a = self.big_gd.get_ranks_from_positions(spos_ac)
        big_partition = AtomPartition(self.big_gd.comm, rank_a)
        return AtomicMatrixDistributor(atom_partition, self.broadcast_comm,
                                       big_partition)


class NullGrid2Grid:
    def __init__(self, aux_gd):
        self.gd = aux_gd
        self.big_gd = aux_gd
        self.enabled = False

    def distribute(self, src_xg, dst_xg=None):
        assert src_xg is dst_xg or dst_xg is None
        return src_xg

    collect = distribute

    def get_matrix_distributor(self, atom_partition, spos_ac=None):
        class NullMatrixDistributor:
            def distribute(self, D_asp):
                return D_asp
            collect = distribute
        return NullMatrixDistributor()

    def new(self, gd, big_gd):
        assert np.all(gd.n_c == big_gd.n_c)
        return NullGrid2Grid(gd)


def grid2grid(comm, gd1, gd2, src_g, dst_g):
    assert np.all(src_g.shape == gd1.n_c)
    assert np.all(dst_g.shape == gd2.n_c)

    #master1_rank = gd1.comm.translate_ranks(comm, [0])[0]
    #master2_rank = gd2.comm.translate_ranks(comm, [0])[0]

    ranks1 = gd1.comm.translate_ranks(comm, np.arange(gd1.comm.size))
    ranks2 = gd2.comm.translate_ranks(comm, np.arange(gd2.comm.size))
    assert (ranks1 >= 0).all(), 'comm not parent of gd1.comm'
    assert (ranks2 >= 0).all(), 'comm not parent of gd2.comm'

    def rank2parpos(gd, rank):
        gdrank = comm.translate_ranks(gd.comm, [rank])[0]
        if gdrank == -1:
            return None
        return gd.get_processor_position_from_rank(gdrank)
    rank2parpos1 = partial(rank2parpos, gd1)
    rank2parpos2 = partial(rank2parpos, gd2)

    general_redistribute(comm,
                         gd1.n_cp, gd2.n_cp,
                         rank2parpos1, rank2parpos2,
                         src_g, dst_g)

def main():
    from gpaw.grid_descriptor import GridDescriptor
    from gpaw.mpi import world
    
    serial = world.new_communicator([world.rank])

    # Genrator which must run on all ranks
    gen = np.random.RandomState(0)

    # This one is just used by master
    gen_serial = np.random.RandomState(17)

    maxsize = 5
    for i in range(1):
        N1_c = gen.randint(1, maxsize, 3)
        N2_c = gen.randint(1, maxsize, 3)
        
        gd1 = GridDescriptor(N1_c, N1_c)
        gd2 = GridDescriptor(N2_c, N2_c)
        serial_gd1 = gd1.new_descriptor(comm=serial)
        serial_gd2 = gd2.new_descriptor(comm=serial)

        a1_serial = serial_gd1.empty()
        a1_serial.flat[:] = gen_serial.rand(a1_serial.size)

        if world.rank == 0:
            print('r0: a1 serial', a1_serial.ravel())

        a1 = gd1.empty()
        a1[:] = -1

        grid2grid(world, serial_gd1, gd1, a1_serial, a1)

        print(world.rank, 'a1 distributed', a1.ravel())
        world.barrier()

        a2 = gd2.zeros()
        a2[:] = -2
        grid2grid(world, gd1, gd2, a1, a2)
        print(world.rank, 'a2 distributed', a2.ravel())
        world.barrier()

        #grid2grid(world, gd2, gd2_serial

        gd1 = GridDescriptor(N1_c, N1_c * 0.2)
        #serialgd = gd2.new_descriptor(

        a1 = gd1.empty()
        a1.flat[:] = gen.rand(a1.size)

        #print a1
        grid2grid(world, gd1, gd2, a1, a2)

        #print a2

if __name__ == '__main__':
    main()



import os
import sys
import time

from ase.units import Bohr
from ase import Atom, Atoms
from ase.parallel import parprint

from gpaw import GPAW
from gpaw.mpi import world 
from gpaw.test import equal
from gpaw.lrtddft import LrTDDFT
from gpaw.lrtddft.excited_state import ExcitedState

txt='-'
txt='/dev/null'

R=0.7 # approx. experimental bond length
a = 3.0
c = 4.0
H2 = Atoms([Atom('H', (a / 2, a / 2, (c - R) / 2)),
            Atom('H', (a / 2, a / 2, (c + R) / 2))],
           cell=(a, a, c))
calc = GPAW(xc='PBE', h=0.25, nbands=3, spinpol=False, txt=txt)
H2.set_calculator(calc)

xc='LDA'
lr = LrTDDFT(calc, xc=xc)

# excited state with forces
accuracy = 0.015
exst = ExcitedState(lr, 0, d=0.01,
        parallel=2,
        txt=sys.stdout,
      )

t0 = time.time()
parprint("########### first call to forces --> calculate")
forces = exst.get_forces(H2)
parprint("time used:", time.time() - t0)
for c in range(2):
    equal(forces[0,c], 0.0, accuracy)
    equal(forces[1,c], 0.0, accuracy)
equal(forces[0, 2] + forces[1, 2], 0.0, accuracy)


parprint("########### second call to potential energy --> just return")
t0 = time.time()
E = exst.get_potential_energy()
parprint("E=", E)
parprint("time used:", time.time() - t0)
t0 = time.time()
E = exst.get_potential_energy(H2)
parprint("E=", E)
parprint("time used:", time.time() - t0)

parprint("########### second call to forces --> just return")
t0 = time.time()
exst.get_forces()
parprint("time used:", time.time() - t0)
t0 = time.time()
exst.get_forces(H2)
parprint("time used:", time.time() - t0)

parprint("###########  moved atoms, call to forces --> calculate")
p = H2.get_positions()
p[1, 1] += 0.1
H2.set_positions(p)

t0 = time.time()
exst.get_forces(H2)
parprint("time used:", time.time() - t0)



# Copyright (C) 2003  CAMP
# Please see the accompanying LICENSE file for further information.

import os
import time
import xml.sax

import numpy as np

import ase
from ase import __version__ as ase_version
import gpaw
from gpaw import __version__ as gpaw_version

try:
    #new style cmr io
    import cmr
    from cmr.base.converter import Converter
    from cmr.io import Flags
    from cmr.tools.functions import create_db_filename as cdbfn
    from cmr.definitions import CALCULATOR_GPAW

    def get_reader(name):
        reader = cmr.read(name, mode=Flags.READ_MODE_ORIGINAL_TYPE)
        return reader
        
    def get_writer():
        return Converter.get_xml_writer(CALCULATOR_GPAW, calculator_version=gpaw_version)
    
    def create_db_filename(param, ext=".db"):
        return cdbfn(param, ext=ext)
    
except:
    #old style cmr io
    import cmr
    from cmr import create_db_filename as cdbfn
    from cmr.io import XMLData
    from cmr.io import READ_DATA
    from cmr.io import EVALUATE
    from cmr.io import WRITE_CONVERT
    from cmr.io import CONVERTION_ORIGINAL
    from cmr.static import CALCULATOR_GPAW
    
    def create_db_filename(param, ext="ignored"):
        return cdbfn()

    def get_reader(name):
        reader = cmr.read(name,
                          read_mode=READ_DATA,
                          evaluation_mode=EVALUATE,
                          convertion_mode=CONVERTION_ORIGINAL)
        return reader
        
    def get_writer():
        data = XMLData()
        data.set_calculator_name(CALCULATOR_GPAW)
        data.set_write_mode(WRITE_CONVERT)
        return data

class Writer:
    """ This class is a wrapper to the db output writer
    and intended to be used with gpaw
    """
    def __init__(self, filename, comm=None):
        self.comm = comm # for possible future use
        self.verbose = False
        self.data = get_writer()
        self.split_array = None #used when array is not filled at once
        self.dimensions = {}
        self.filename = filename
        self.cmr_params = {}

        uname = os.uname()
        self.data['user']=os.getenv('USER', '???')
        self.data['date']=time.asctime()

        self.data['architecture']=uname[4]
        self.data['ase_dir']=os.path.dirname(ase.__file__)
        self.data['ase_version']=ase_version
        self.data['numpy_dir']=os.path.dirname(np.__file__)
        self.data['gpaw_dir']=os.path.dirname(gpaw.__file__)
        self.data["db_calculator_version"] = gpaw_version
        self.data['calculator']="gpaw"
        self.data['location']=uname[1]

        
    def dimension(self, name, value):
        self.dimensions[name]=value
        self.data[name]=value
        if self.verbose:
            print("dimension: ", name, value)

    def __setitem__(self, name, value):
        """ sets the value of a variable in the db-file. Note that only
        values that were defined in the cmr-schema are written.
        (User defined values have to be added with set_user_variable(name, value)
        IMPORTANT: CMR does not support None values for numbers, therefore all variables
        with value None are ignored."""
        if self.verbose:
            print("name value:", name, value)
        if name == "GridSpacing" and value == "None":
            return
        if not value is None:
            self.data[name]=value
        
    def _get_dimension(self, array):
        """retrieves the dimension of a multidimensional array
        by using then len() function until fail
        """
        indent = ""
        measured_dimensions = []
        while True:
            try:
                measured_dimensions.append(len(array))
                if self.verbose:
                    print(indent+"Length:", len(array))
                    indent += " "
                array = array[0]
            except IndexError:
                break
            except TypeError:
                break
        return measured_dimensions

    def _close_array(self):
        """if an array is filled with fill then we don't know
        the exact end of the array. Therefore we check before
        adding a new variable if the end was reached."""
        if self.split_array is None:
            return
        
        measured_dimensions = self._get_dimension(self.split_array[2])
        if self.verbose:
            print("Dimensions:         ", self.split_array[1])
            print("Mesured Dimensions: ", measured_dimensions)
            
        #make the array fit (only fixes the 1st dimension)
        res = np.array(self.split_array[2]).reshape(self.split_array[1])
        self.data[self.split_array[0]] = res

    def add(self, name, shape, array=None, dtype=None, units=None,
            parallel=False, write=True):
        self._close_array()
        if self.verbose:
            print("add:", name, shape, array, dtype, units)
        if array is None:
            dimension = []
            for a in shape:
                dimension.append(self.dimensions[a])
            self.split_array = (name, dimension, [])
        else:
            self.data[name]=array

    def fill(self, array, *indices, **kwargs):
        if self.verbose:
            print("fill (", len(array),"):", array)
        self.split_array[2].append(array)

    def set_db_copy_settings(self, make_db_copy, private):
        """deprecated: This method is going to be removed"""
        print("Warning: gpaw.cmr.readwriter.set_db_copy_settings is deprecated.")
        print("Please update gpaw and CMR")
        pass
        #if self.verbose:
        #    print "set_db_copy_settings", make_db_copy, private
        #self.db_copy=make_db_copy
        #self.private=private
        #self.data.set_db_copy_settings(make_db_copy, private)

    def write(self, string, db, private, **kwargs):
        print("Warning: gpaw.cmr.readwriter.write is deprecated.")
        print("Please update gpaw and CMR")
        pass
        #if self.verbose:
        #    print "write():", string
        #self.data.write(string, db, private, **kwargs)

    def write_additional_db_params(self, cmr_params):
        """writes the user variables and also sets the write attributes for
        the output file"""
        self.cmr_params = cmr_params.copy()
        #cmr.set_params_to_xml_data(self.data, cmr_params)
        
    def close(self):
        if self.verbose:
            print("close()")
        self._close_array()
        if "ase_atoms_var" in self.cmr_params:
            ase_vars = self.cmr_params["ase_atoms_var"]
            for key in ase_vars:
                self.data.set_user_variable(key, ase_vars[key])
            self.cmr_params.pop("ase_atoms_var")
        if self.filename==".db" or self.filename==".cmr":
            # Note:
            #      .cmr files can currently not be uploaded to the database therefore
            #      it defaults to .db until supported
            self.cmr_params["output"]=create_db_filename(self.data, ext=".db")
        else:
            self.cmr_params["output"]=self.filename
        try:
            cmr.runtime.pause_ase_barriers(True)
            self.data.write(self.cmr_params, ase_barrier=False)
        except TypeError:
            # for compatibility with older CMR versions:
            self.data.write(self.cmr_params)
        cmr.runtime.pause_ase_barriers(False)
        return [self.data.get_hash()]


class Reader:
    """ This class allows gpaw to access
    to read a db-file
    """
    def __init__(self, name, comm):
        self.verbose = False
        self.reader = self.parameters = get_reader(name)

    def dimension(self, name):
        return self.reader[name]
    
    def __getitem__(self, name):
        if name=='version' and 'version' not in self.reader \
            and 'db_calculator_version' in self.reader:
                return self.reader['db_calculator_version']
        return self.reader[name]

    def has_array(self, name):
        return name in self.reader
    
    def get(self, name, *indices, **kwargs):
        if self.verbose:
            print("incides", indices)
        result = self.reader[name]
        if indices!=():
            for a in indices:
                result = result[a]
            return result
        
        #gpaw wants expressions evaluated
        if isinstance(result, str) or isinstance(result, unicode):
            try:
                if self.verbose:
                    print("Converting ", result)
                result = eval(result, {})
            except (SyntaxError, NameError):
                pass
        return result
    
    def get_reference(self, name, indices, length=None):
        result = self.reader[name]
        if indices!=():
            for a in indices:
                result = result[a]
            return result
    
    def get_file_object(self, name, indices):
        result = self.reader.retrieve_file(name)
        if indices!=():
            for a in indices:
                result = result[a]
        return result
        

    def get_parameters(self):
        return self.reader.keys()

    def close(self):
        pass





import glob
import os
import subprocess

from ase.utils.build_web_page import build, git_pull, main


def build_gpaw_web_page(force_build):
    changes = git_pull('gpaw')
    if force_build or changes:
        os.chdir('ase')
        subprocess.check_call('python setup.py install --home=..', shell=True)
        os.chdir('..')
        subprocess.check_call(
            'wget --no-check-certificate --quiet --timestamping '
            'http://wiki.fysik.dtu.dk/gpaw-files/'
            'gpaw-setups-latest.tar.gz', shell=True)
        subprocess.check_call(
            'tar xf gpaw-setups-latest.tar.gz', shell=True)
        path = os.path.abspath(glob.glob('gpaw-setups-[0-9]*')[0])
        build(True, 'gpaw', 'GPAW_SETUP_PATH=' + path)
        
        
if __name__ == '__main__':
    main(build_gpaw_web_page)



from __future__ import print_function
import numpy as np
from ase.lattice import bulk
from gpaw import GPAW, PW


si = bulk('Si')
k = 3
si.calc = GPAW(mode=PW(250),
               xc='PBE',
               kpts=(k, k, k),
               convergence={'energy': 1e-8},
               txt='si.txt')

si.set_cell(np.dot(si.cell,
                   [[1.02, 0, 0.03],
                    [0, 0.99, -0.02],
                    [0.2, -0.01, 1.03]]),
            scale_atoms=True)

sigma_vv = si.get_stress(voigt=False)
print(sigma_vv)

deps = 1e-5
cell = si.cell.copy()
for v in range(3):
    x = np.eye(3)
    x[v, v] += deps
    si.set_cell(np.dot(cell, x), scale_atoms=True)
    ep = si.calc.get_potential_energy(si, force_consistent=True)
    x[v, v] -= 2 * deps
    si.set_cell(np.dot(cell, x), scale_atoms=True)
    em = si.calc.get_potential_energy(si, force_consistent=True)
    s = (ep - em) / 2 / deps / si.get_volume()
    print((v, s, abs(s - sigma_vv[v, v])))
    assert abs(s - sigma_vv[v, v]) < 1e-4
for v1 in range(3):
    v2 = (v1 + 1) % 3
    x = np.eye(3)
    x[v1, v2] = deps
    x[v2, v1] = deps
    si.set_cell(np.dot(cell, x), scale_atoms=True)
    ep = si.calc.get_potential_energy(si, force_consistent=True)
    x[v1, v2] = -deps
    x[v2, v1] = -deps
    si.set_cell(np.dot(cell, x), scale_atoms=True)
    em = si.calc.get_potential_energy(si, force_consistent=True)
    s = (ep - em) / deps / 4 / si.get_volume()
    print((v1, v2, s, abs(s - sigma_vv[v1, v2])))
    assert abs(s - sigma_vv[v1, v2]) < 2e-4



import sys

from ase.data.g2_1 import data
from ase.data.g2_1 import atom_names
from ase.data.g2_1 import molecule_names
from ase.tasks.main import run
from ase.tasks.molecule import MoleculeTask

from ase.structure import molecule

from gpaw.test.big.scf.analyse import rundefs

class Collection:

    def __init__(self, data, names, cell):
        self.data = data
        self.names = names
        self.cell = cell

    def __getitem__(self, name):
        return self.create_item(name)

    def keys(self):
        return self.names

    def create_item(self, name):
        m = molecule(name, data=self.data)
        m.set_cell(self.cell)
        m.set_pbc((0, 0, 0))
        m.center()
        return m

class Task(MoleculeTask):
    def __init__(self, **kwargs):
        MoleculeTask.__init__(self, **kwargs)

    def calculate(self, name, atoms):
        data = MoleculeTask.calculate(self, name, atoms)
        try:
            steps = atoms.get_calculator().get_number_of_iterations()
        except (AttributeError, NotImplemented):
            steps = None
        data['calculator steps'] = steps
        return data

from gpaw.mixer import Mixer, MixerSum, MixerSum2, MixerDif
from gpaw.mixer import BroydenMixer, BroydenMixerSum
from gpaw.factory import GPAWFactory

if len(sys.argv) == 1:
    run = None
else:
    run = sys.argv[1]

cell = (12.01, 12.02, 12.03)

tag = 'scf_g2_1_pbe0_fd'

if run is not None:
    tag += '_%s' % run

calcopts = {
    'mode': 'fd',
    'xc': 'PBE0',
    'width': 0.0,
    'fixmom': True,
    # allow other mixers
    'spinpol': True,
    # allow for long SCFs
    'maxiter': 700,
    }

if run.startswith('bands'):
    nbands = run[len('bands'):len('bands') + 2]
    if nbands == '00':
        calcopts.update({'nbands': None})
    else:
        calcopts.update({'nbands': - int(nbands)})
if run.startswith('cgbands'):
    nbands = run[len('cgbands'):len('cgbands') + 2]
    if nbands == '00':
        calcopts.update({'nbands': None})
    else:
        calcopts.update({'nbands': - int(nbands)})
if run.startswith('dzpbands'):
    nbands = run[len('dzpbands'):len('dzpbands') + 2]
    if nbands == '00':
        calcopts.update({'nbands': None})
    else:
        calcopts.update({'nbands': - int(nbands)})
if run.startswith('szpdzp'):
    calcopts.update({'basis': 'szp(dzp)'})
if run.startswith('szdzp'):
    calcopts.update({'basis': 'sz(dzp)'})
if run.startswith('dzp'):
    calcopts.update({'basis': 'dzp'})
if 'cg' in run:
    calcopts.update({'eigensolver': 'cg'})
else:
    calcopts.update({'eigensolver': 'rmm-diis'})
if run.startswith('cgdzp'):
    calcopts.update({'basis': 'dzp'})
calcopts.update({'mixer': eval(rundefs[run])})

calcfactory = GPAWFactory(**calcopts)

collection = Collection(data, atom_names + molecule_names, cell)

task = Task(calcfactory=calcfactory,
            tag=tag,
            use_lock_files=True,
            collection=collection,
            cell=cell,
            )
keys = collection.keys()
for m in ['Na2', 'NaCl']:  # those seem to need cg
    keys.remove(m)
task.run(keys)
